''' Pictorial Style Elements '''
from __future__ import annotations
from typing import Optional
import math

from ..backends.svgunits import parse_size_to_px, PX_PER_IN
from ..segments import (Segment,
                        SegmentPath,
                        SegmentPoly,
                        SegmentArc,
                        SegmentCircle,
                        SegmentText)
from ..elements import Element, Element2Term


# Scale factors in terms of drawing units
# INCH chosen to scale the dimensioned pictorial elements to 
# approximately the right size relative to other schemdraw elements
INCH = 3.5  # Drawing units per inch
MILLIMETER = INCH / 25.4   # Drawing units per millimeter
PINSPACING = 0.1 * INCH  # Spacing between breadboard pins

HOUSING_COLOR = '#333'  # Color of ICs, Diodes, etc.
BORDER_COLOR = '#777'   # Border around housings
LEAD_COLOR = '#A0A0A0'  # Color of leads/metal


def parse_size_to_units(value: str) -> float:
    ''' Parse a SVG size string (sunh as '2in') to drawing units '''
    value_px = parse_size_to_px(value)
    value_units = value_px / PX_PER_IN * INCH
    return value_units


def resistor_colors(value: float,
                    tolerance: Optional[float] = None
                    ) -> tuple[str, str, str, Optional[str]]:
    ''' Determine Resistor color bands given the value (ohms) and tolerance (in percent)'''
    colors = {
        0: 'black',
        1: 'brown',
        2: 'red',
        3: 'orange',
        4: 'yellow',
        5: 'green',
        6: 'blue',
        7: 'violet',
        8: 'gray',
        9: 'white',
        -1: 'gold',
        -2: 'silver',
        -3: 'pink'}
    
    tol = {  # in percent
        1: 'brown',
        2: 'red',
        .05: 'orange',
        .02: 'yellow',
        .5: 'green',
        .25: 'blue',
        .1: 'violet',
        .01: 'gray',
        5: 'gold',
        10: 'silver',
        None: None}

    mult = int(math.log10(value))-1  # Third band/multiplier
    base = value / 10**mult
    if base < 10:
        mult-=1
        base = value / 10**mult
    band1 = int(base / 10)
    band2 = int(base % 10)
    return colors[band1], colors[band2], colors[mult], tol.get(tolerance, None)


class ElementPictorial(Element2Term):
    ''' This class gives pictorial 2-term elements silver lead extensions '''
    _element_defaults = {
        'leadcolor': LEAD_COLOR,
        'lw': 3,
    }


class Resistor(ElementPictorial):
    ''' Carbon-film 1/4 Watt Resistor

        Args:
            value: Resistance value, used to set color bands
            tolerance: Tolerance value for 4th color band

        Note: Color bands will be closest possible value that can be represented
            with 3 bands
    '''
    _element_defaults = {
        'fill': 'blanchedalmond'
    }
    def __init__(self, value: float = 1000, tolerance: Optional[float] = None, **kwargs):
        super().__init__(**kwargs)
        # Spec dimensions are w = 6.3mm, h = 2.3mm. This is close enough.
        w = PINSPACING    # width of straight part (in direction of current)
        h = PINSPACING/2  # height of rounded part
        h2 = h * 0.9      # height of straight part

        # for 2-term endpoints
        self.segments.append(Segment(((0, 0), (math.nan, math.nan), (h*3+w, 0))))
        self.segments.append(SegmentPath(
            ('M', (0, 0),
             'Q', (0, h), (h, h),
             'Q', (h*1.25, h), (h*1.5, h2),
             'L', (h*1.5+w, h2),
             'Q', (h*1.75+w, h), (h*2+w, h),
             'Q', (h*3+w, h), (h*3+w, 0),
             'Q', (h*3+w, -h), (h*2+w, -h),
             'Q', (h*1.75+w, -h), (h*1.5+w, -h2),
             'L', (h*1.5, -h2),
             'Q', (h*1.25, -h), (h, -h),
             'Q', (0, -h), (0, 0),
             'Z'),
            color='black',
            lw=1))

        # Color bands
        stripe1 = .16
        stripe2 = .32
        stripe3 = .48
        stripe4 = .68
        stripew = .08
        c1, c2, c3, ct = resistor_colors(value, tolerance)
        self.segments.append(SegmentPoly(((stripe1, h2), (stripe1, -h2),
                                          (stripe1+stripew, -h2), (stripe1+stripew, h2)),
                                          fill=c1, lw=.1, zorder=2))

        self.segments.append(SegmentPoly(((stripe2, h2), (stripe2, -h2),
                                          (stripe2+stripew, -h2), (stripe2+stripew, h2)),
                                          fill=c2, lw=.1, zorder=2))

        self.segments.append(SegmentPoly(((stripe3, h2), (stripe3, -h2),
                                          (stripe3+stripew, -h2), (stripe3+stripew, h2)),
                                          fill=c3, lw=.1, zorder=2))

        if ct:
            self.segments.append(SegmentPoly(((stripe4, h2), (stripe4, -h2),
                                              (stripe4+stripew, -h2), (stripe4+stripew, h2)),
                                              fill=ct, lw=.1, zorder=2))


class Diode(ElementPictorial):
    ''' Diode in DO-204 package '''
    _element_defaults = {
        'fill': HOUSING_COLOR,
        'stripe_color': 'silver'
    }
    def __init__(self, *, stripe_color: Optional[str] = None, **kwargs):
        super().__init__(**kwargs)
        h = PINSPACING/2
        w = PINSPACING * 2.5
        stripex = w * .78
        stripew = .08
        stripe_color = self.params['stripe_color']

        self.segments.append(Segment(((0, 0), (math.nan, math.nan), (w, 0))))  # 2-term endpoints
        self.segments.append(SegmentPoly(((0, h), (w, h), (w, -h), (0, -h)), closed=True, lw=1, color='black', zorder=2))
        self.segments.append(SegmentPoly(((stripex, h), (stripex+stripew, h), (stripex+stripew, -h), (stripex, -h)),
                                          closed=True, lw=1, color='none', fill=stripe_color, zorder=2))


class LED(Element):
    ''' Light Emitting Diode Use .fill() to set the color.
    
        Suggested fill colors for common LEDs:
            * red: #dd4433,
            * orange: #efa207
            * yellow: #f1de0f
            * green: #4be317
            * blue: #3892bc
            * white/clear: #e5e5e5
    '''
    _element_defaults = {
        'lead_length': PINSPACING*5,
        'fill': '#dd4433',
        'theta': 0,
    }
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        # From datasheet
        width = .197 * INCH
        height = .338 * INCH

        # Leads
        lead_h = self.params['lead_length']  # Lead to bottom of LED
        ref = PINSPACING * 5
        self.segments.append(
            Segment(((0, 0), (0, lead_h)), color=LEAD_COLOR, lw=4))
        self.segments.append(
            Segment(((PINSPACING, 0), (PINSPACING, lead_h-ref*.3),
                     (PINSPACING*.6, lead_h-ref*.2), (PINSPACING*.6, lead_h)),
                     color=LEAD_COLOR, lw=4))

        top_r = width * .8
        left = PINSPACING/2 - width/2
        center = PINSPACING/2
        top = lead_h+height

        # Case
        self.segments.append(
            SegmentPath(('M', (left, lead_h),
                         'Q', (center, lead_h-ref*.2), (left+width, lead_h),
                         'L', (left+width, top-top_r),
                         'C', (left+width, top), (left, top), (left, top-top_r),
                         'Z'), color='none'))  # color set by fill

        # Accents - use white with transparency to be color-agnostic
        # NOTE: linewidth won't scale with zoom
        base_h = .039 * INCH
        self.segments.append(
            SegmentPath(('M', (left, lead_h+base_h/2),
                         'Q', (center*.5, lead_h-ref*.16+base_h), (left+width*.7, lead_h-ref*.03),
                         'L', (left+width+.0*INCH, lead_h+base_h/2)),
                        color='#FFFFFF55', lw=2))

        reflectionx = left+.03*INCH
        self.segments.append(
            SegmentPath(('M', (reflectionx, lead_h+base_h*2),
                         'L', (reflectionx, top-top_r),
                         'Q', (reflectionx, top-top_r+.04*INCH),
                              (reflectionx+.04*INCH, top-top_r+.07*INCH)),
                        color='#FFFFFF55', lw=3))


class LEDOrange(LED):
    ''' Orange Light Emitting Diode '''
    _element_defaults = {
        'fill': '#efa207'
    }


class LEDYellow(LED):
    ''' Yellow Light Emitting Diode '''
    _element_defaults = {
        'fill': '#f1de0f'
    }


class LEDGreen(LED):
    ''' Green Light Emitting Diode '''
    _element_defaults = {
        'fill': '#4be317'
    }


class LEDBlue(LED):
    ''' Blue Light Emitting Diode '''
    _element_defaults = {
        'fill': '#3892bc'
    }


class LEDWhite(LED):
    ''' White Light Emitting Diode '''
    _element_defaults = {
        'fill': '#e5e5e5'
    }



class CapacitorElectrolytic(Element):
    ''' Electrolytic capacitor

        Note: Use .fill() to change color.

        Args:
            cap_color: Color of the metal cap on top
            stripe_color: Color of the polarity stripe
    '''
    _element_defaults = {
        'fill': '#303b91',
        'lead_length': 0.2*INCH,
        'cap_color': '#CCC',
        'stripe_color': '#dff1f6',
        'theta': 0,
    }
    def __init__(self, *, lead_length: Optional[float] = None,
                 cap_color: Optional[str] = None,
                 stripe_color: Optional[str] = None, **kwargs):
        super().__init__(**kwargs)

        width = PINSPACING*1.7
        height = .3*INCH

        # Leads
        lead_h = self.params['lead_length']  # Lead to bottom of LED
        self.segments.append(
            Segment(((0, 0), (0, lead_h)), color=LEAD_COLOR, lw=4))
        self.segments.append(
            Segment(((PINSPACING, 0), (PINSPACING, lead_h)), color=LEAD_COLOR, lw=4))

        ref = PINSPACING*2
        top_r = ref*.3
        left = PINSPACING/2 - width/2
        center = PINSPACING/2
        top = lead_h + height

        # Case
        self.segments.append(
            SegmentPath(('M', (left, lead_h),
                         'Q', (center, lead_h - ref*.3), (left+width, lead_h),
                         'L', (left+width, top-top_r),
                         'Q', (center, top), (left, top-top_r),
                         'Z'), color='none'))  # color set by fill

        # Cap
        self.segments.append(
            SegmentArc((center, top-top_r-.01*INCH), width-.03*INCH, top_r*.9, 0, 360,
                       fill=self.params['cap_color'], color='none', lw=.5, zorder=2))

        # Bottom accent
        base_h = .039*INCH / 2
        self.segments.append(
            SegmentPath(('M', (left, lead_h+base_h),
                         'Q', (center, lead_h-ref*.1), (left+width, lead_h+base_h)),
                        color='#FFFFFF99', fill='none', lw=1.5))

        # Polarity stripe
        stripex = left + 0.03*INCH
        self.segments.append(
            SegmentPath(('M', (stripex, lead_h+base_h),
                         'L', (stripex, top-top_r*2),),
                        color=self.params['stripe_color'], fill='none', lw=3))


class CapacitorMylar(Element):
    ''' Mylar Capacitor '''
    _element_defaults = {
        'fill': '#8c2510',
        'lead_length': 0.4*INCH,
        'theta': 0,
    }
    def __init__(self, *, lead_length: Optional[float] = None, **kwargs):
        super().__init__(**kwargs)
        width = PINSPACING*1.75
        height = width*.75
        center = PINSPACING/2
        left = center-width/2

        # Leads
        lead_h = self.params['lead_length']
        self.segments.append(
            Segment(((0, 0), (0, lead_h)), color=LEAD_COLOR, lw=4))
        self.segments.append(
            Segment(((PINSPACING, 0), (PINSPACING, lead_h)), color=LEAD_COLOR, lw=4))

        bulge = .02*INCH
        self.segments.append(
            SegmentPath(('M', (left, lead_h),
                         'Q', (left-bulge, lead_h+height/2), (left, lead_h+height),
                         'Q', (center, lead_h+height+bulge), (left+width, lead_h+height),
                         'Q', (left+width+bulge, lead_h+height/2), (left+width, lead_h),
                         'Q', (center, lead_h+bulge/2), (left, lead_h),
                         'Z'),
                        color='none'))

        highlighty = lead_h+height-.025*INCH
        self.segments.append(
            Segment(((-.01*INCH, highlighty), (PINSPACING*.8, highlighty)),
                    color='#FFFFFF44', lw=3))


class CapacitorCeramic(Element):
    ''' Ceramic Disc Capacitor '''
    _element_defaults = {
        'fill': '#e2800b',
        'radius': 0.1*INCH,
        'lead_length': 0.4*INCH,
        'theta': 0,
    }
    def __init__(self, *, radius: Optional[float] = None,
                 lead_length: Optional[float] = None, **kwargs):
        super().__init__(**kwargs)
        rad = self.params['radius']
        lead_h = self.params['lead_length']
        center = PINSPACING/2, lead_h+rad-.05*INCH

        # Leads
        self.segments.append(
            Segment(((0, 0), (0, lead_h)), color=LEAD_COLOR, lw=4))
        self.segments.append(
            Segment(((PINSPACING, 0), (PINSPACING, lead_h)), color=LEAD_COLOR, lw=4))

        self.segments.append(
            SegmentCircle(center, rad, color='none', zorder=2))

        self.segments.append(
            SegmentArc(center, rad*2*.7, rad*2*.7, 100, 200,
                       color='#FFFFFF44', zorder=2))


class TO92(Element):
    ''' TO92 Transistor Package, with pins bent out to 0.1 inch breadboard spacing '''
    _element_defaults = {
        'theta': 0,
    }
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        # Dimensions from TO92 spec
        width = .175*INCH
        height = .170*INCH*.9  # .170 is standard height, but *.9 due to perspective
        topheight = PINSPACING*1.2

        # (0, 0) is left pin. case_x, y is lower left of the case
        case_y = PINSPACING*1.3
        case_x = PINSPACING-width/2
        center = PINSPACING
        self.segments.append(
            Segment(((0, 0), (0, PINSPACING*.8), (PINSPACING/2, case_y)),
                    color=LEAD_COLOR, lw=4))
        self.segments.append(
            Segment(((center, 0), (center, case_y)),
                    color=LEAD_COLOR, lw=4))
        self.segments.append(
            Segment(((PINSPACING*2, 0), (PINSPACING*2, PINSPACING*.8), (3*PINSPACING/2, case_y)),
                    color=LEAD_COLOR, lw=4))

        self.segments.append(
            SegmentPoly(((case_x, case_y), (case_x+width, case_y),
                         (case_x+width, case_y+height), (case_x, case_y+height)),
                        fill=HOUSING_COLOR, lw=1, color=BORDER_COLOR, zorder=2))
        self.segments.append(
            SegmentArc((center, case_y+height), width, topheight, 0, 180,
                       fill=HOUSING_COLOR, lw=1, color=BORDER_COLOR, zorder=2))

        # using "pin" since emitter/base/collector aren't standardized
        self.anchors['pin1'] = (0, 0)
        self.anchors['pin2'] = (PINSPACING, 0)
        self.anchors['pin3'] = (PINSPACING*2, 0)
        self.anchors['case_center'] = (center, case_y+height/2)
        self.elmparams['lblloc'] = 'case_center'
        self.elmparams['lblalign'] = ('center', 'center')


class DIP(Element):
    ''' Dual-Inline-Package IC

        Args:
            npins: Total number of pins
            wide: Use wide (0.6 inch) package instead of narrow
                (0.3 inch) package
    '''
    _element_defaults = {
        'theta': 0
    }
    def __init__(self, npins: int = 8, wide: bool = False, **kwargs):
        super().__init__(**kwargs)
        if npins % 2 != 0:
            npins += 1  # Can't have odd number, just round it up

        # Align Pin1 with (0, 0)
        # standard DIP width is 0.3 or 0.6 inch
        # width is from pin1 to pin8
        width = 3*PINSPACING if not wide else 6*PINSPACING

        pady = PINSPACING*.25  # padding to add above top pin and below bottom pin
        padx = PINSPACING*.25  # padding to remove left/right
        height = (npins//2-1) * PINSPACING #+ pady*2
        center = (width)/2
        notch_r = PINSPACING*.75
        body_left = padx
        body_right = width-padx
        leadw = 5
        
        for n in range(npins//2):
            self.segments.append(Segment(((0, -n*PINSPACING), (body_left, -n*PINSPACING)), color=LEAD_COLOR, lw=leadw, zorder=1))
            self.segments.append(Segment(((body_right, -n*PINSPACING), (body_right+padx, -n*PINSPACING)), color=LEAD_COLOR, lw=leadw, zorder=1))
            self.anchors[f'pin{n+1}'] = (0, -n*PINSPACING)
            self.anchors[f'pin{npins-n}'] = (body_right+padx, -n*PINSPACING)

        self.segments.append(SegmentPoly(((padx, pady), (body_right, pady),
                                          (body_right, -pady-height), (padx, -pady-height)), fill=HOUSING_COLOR, lw=1, color='#444'))
        self.segments.append(SegmentArc((center, pady), notch_r, notch_r, 180, 0, lw=1, color='#777', fill='#555'))

        self.elmparams['lblloc'] = 'center'
        self.elmparams['lblrotate'] = True
        self.elmparams['lblofst'] = 0


class Breadboard(Element):
    ''' Solderless Breadboard, 400-point (30 rows)

        Use .color() to set border color, and .fill() to set fill color.
    '''
    _element_defaults = {
        'color': '#DDD',  # Border
        'fill': '#F8F8F5',
        'shadow_color': '#E5E5E5',
        'text_color': '#666',
    }
    def __init__(self, *,
                 shadow_color: Optional[str] = None,
                 text_color: Optional[str] = None,
                 **kwargs):
        super().__init__(**kwargs)

        inner_radius = .02
        outer_radius = .1
        shadow_color = self.params.get('shadow_color')
        text_color = self.params.get('text_color')

        def hole(x, y):
            self.segments.append(SegmentPoly(((x-outer_radius, y+outer_radius), (x+outer_radius, y+outer_radius),
                                                (x+outer_radius, y-outer_radius)),
                                                fill=True, color=shadow_color, lw=2))
            self.segments.append(SegmentPoly(((x-inner_radius, y-inner_radius), (x-inner_radius, y+inner_radius),
                                                (x+inner_radius, y+inner_radius), (x+inner_radius, y-inner_radius)),
                                                fill=True, color=HOUSING_COLOR, lw=2))

        nrows = 30
        ncols = 5

        # Dimensionss from breadboard spec sheets
        strip_w = .37 * INCH  # Edge/power strips are 0.37 inch
        edgepad = (strip_w - PINSPACING)/2  # distance from edge to center of first pin in the strip
        width = 2.10 * INCH  # Total width
        height = 3.3 * INCH  # Total height
        top = 1.5 * edgepad  # Offset the top/left so that (0, 0) is top of left power strip
        left = -edgepad
        right = left+width

        # Frame - inherit color from element
        self.segments.append(
            SegmentPoly(((left, top), (right, top),
                         (right, top-height), (left, top-height)), lw=2))

        x = y = 0.
        # Left power strip holes
        for col in range(2):
            for row in range(nrows):
                if row % 6 == 0: continue
                xy1 = x, y-row*PINSPACING
                xy2 = x+PINSPACING, y-row*PINSPACING 
                hole(*xy1)
                hole(*xy2)
                self.anchors[f'L1_{row}'] = xy1
                self.anchors[f'L2_{row}'] = xy2

        # Right power strip holes
        x = right - edgepad
        for col in range(2):
            for row in range(nrows):
                if row % 6 == 0: continue
                xy1 = x, y-row*PINSPACING
                xy2 = x-PINSPACING, y-row*PINSPACING
                hole(*xy1)
                hole(*xy2)
                self.anchors[f'R1_{row}'] = xy2
                self.anchors[f'R2_{row}'] = xy1

        # Inner Rows
        x = strip_w
        for col in range(ncols):
            colname = chr(ord('A')+col)
            self.segments.append(SegmentText((x+col*PINSPACING, y+PINSPACING), colname,
                                             fontsize=8, rotation_global=False, align=('center', 'center'),
                                             color=text_color, zorder=1))
            for row in range(nrows):
                xy = x+col*PINSPACING, y-row*PINSPACING
                hole(*xy)
                self.anchors[f'{colname}{row+1}'] = xy
        
        for col in range(ncols):
            colname = chr(ord('F')+col)
            self.segments.append(SegmentText((x+ PINSPACING*7+col*PINSPACING, y+PINSPACING), colname,
                                             fontsize=8, rotation_global=False, align=('center', 'center'),
                                             color=text_color, zorder=1))
            for row in range(nrows):
                xy = x+ PINSPACING*7 + col*PINSPACING, y-row*PINSPACING
                hole(*xy)
                self.anchors[f'{colname}{row+1}'] = xy
        
        # Number Labels
        for row in range(0, nrows):
            self.segments.append(SegmentText((x-PINSPACING, y-row*PINSPACING-.04), str(row+1), fontsize=8,
                                             rotation_global=False, align=('center', 'bottom'), color=text_color, zorder=1))
            self.segments.append(SegmentText((x+PINSPACING*12, y-row*PINSPACING-.04), str(row+1), fontsize=8,
                                             rotation_global=False, align=('center', 'bottom'), color=text_color, zorder=1))


        # Vertical dividing lines
        self.segments.append(Segment(((left+edgepad*.4, top-PINSPACING*2), (left+edgepad*.4, top-height+PINSPACING)), color='red', lw=.5))
        self.segments.append(Segment(((left+strip_w-edgepad*.4, top-PINSPACING*2), (left+strip_w-edgepad*.4, top-height+PINSPACING)), color='blue', lw=.5))
        self.segments.append(Segment(((right-edgepad*.4, top-PINSPACING*2), (right-edgepad*.4, top-height+PINSPACING)), color='blue', lw=.5))
        self.segments.append(Segment(((right-strip_w+edgepad*.4, top-PINSPACING*2), (right-strip_w+edgepad*.4, top-height+PINSPACING)), color='red', lw=.5))



''' Connectors and bus lines '''

from __future__ import annotations
from typing import Optional, Sequence
import warnings

from ..segments import Segment, SegmentText, SegmentCircle, SegmentPoly
from ..elements import Element, Line
from .twoterm import resheight, gap
from ..types import Point, XY, Valign, HeaderStyle, HeaderNumbering


class OrthoLines(Element):
    ''' Orthogonal multiline connectors

        Use `at()` and `to()` methods to specify starting and
        ending location of OrthoLines.

        The default lines are spaced to provide connection to
        pins with default spacing on Ic element or connector
        such as a Header.

        Args:
            n: Number of parallel lines
            dy: Distance between parallel lines [default: 0.6]
            xstart: Fractional distance (0-1) to start vertical
                portion of first ortholine
    '''
    _element_defaults = {
        'dy': 0.6,
    }
    def __init__(self, *,
                 n: int = 1,
                 dy: Optional[float] = None,
                 xstart: Optional[float] = None,
                 arrow: Optional[str] = None,
                 **kwargs):
        super().__init__(**kwargs)
        self._userparams.setdefault('to', (1, 1))

    def to(self, xy: XY) -> 'Element':
        ''' Specify ending position of OrthoLines '''
        self._userparams['to'] = xy
        return self

    def delta(self, dx: float = 0, dy: float = 0) -> 'Element':
        ''' Specify ending position relative to start position '''
        self._userparams['delta'] = Point((dx, dy))
        return self

    def _place(self, dwgxy: XY, dwgtheta: float, **dwgparams) -> tuple[Point, float]:
        ''' Calculate absolute placement of Element '''
        self._dwgparams = dwgparams
        if not self._positioned:
            self._position()

        self.params['theta'] = 0
        xy: XY = self.params.get('at', dwgxy)
        to: XY = self.params.get('to', dwgxy)
        delta = self.params.get('delta', None)
        n = self.params.get('n', 1)
        ndy = self.params.get('dy', .6)
        xstart = self.params.get('xstart', None)
        arrow = self.params.get('arrow', None)
        if delta is not None:
            dx, dy = delta
        else:
            dx = to[0] - xy[0]
            dy = to[1] - xy[1]

        if abs(dy) < .05:
            for i in range(n):
                y = -i*ndy
                self.segments.append(Segment([(0, y), (dx, y)]))
        else:
            # x0 is first line to go up
            if xstart is not None:
                if dy > 0:
                    x0 = dx*xstart
                else:
                    x0 = dx*xstart - ndy - ndy*(n-1)*xstart
                    # xstart=0 --> -ndy; xstart=1 --> dx-ndy*n
            elif dx > 0:
                x0 = dx/2 - (ndy*(n-1)/2)
            else:
                x0 = dx/2 + (ndy*(n-1)/2)

            for i in range(n):
                y = -i*ndy
                if dy > 0:
                    x = x0 + ndy*i if dx > 0 else x0 - ndy*i
                else:
                    x = x0 + (n-i)*ndy if dx > 0 else x0 - (n-i)*ndy
                self.segments.append(Segment([(0, y), (x, y), (x, y+dy), (dx, y+dy)], arrow=arrow))
        return super()._place(dwgxy, dwgtheta, **dwgparams)


class RightLines(Element):
    ''' Right-angle multi-line connectors

        Use `at()` and `to()` methods to specify starting and ending
        location.

        The default lines are spaced to provide connection to
        pins with default spacing on Ic element or connector
        such as a Header.

        Args:
            n: Number of parallel lines
            dy: Distance between parallel lines [default: 0.6]
    '''
    _element_defaults = {
        'dy': 0.6,
        'to': (3, -2),
        'theta': 0,
        'n': 1
    }
    def __init__(self, *,
                 n: int = 1,
                 dy: Optional[float] = None,
                 arrow: Optional[str] = None,
                 **kwargs):
        super().__init__(**kwargs)

    def to(self, xy: XY) -> 'Element':
        ''' Specify ending position of OrthoLines '''
        self._userparams['to'] = xy
        return self

    def delta(self, dx: float = 0, dy: float = 0) -> 'Element':
        ''' Specify ending position relative to start position '''
        self._userparams['delta'] = Point((dx, dy))
        return self

    def _place(self, dwgxy: XY, dwgtheta: float, **dwgparams) -> tuple[Point, float]:
        ''' Calculate absolute placement of Element '''
        self._dwgparams = dwgparams
        if not self._positioned:
            self._position()

        xy = self.params.get('at', dwgxy)
        delta = self.params.get('delta', None)
        to = self.params['to']
        n = self.params['n']
        ndy = self.params['dy']
        arrow = self.params.get('arrow', None)
        if delta is not None:
            dx, dy = delta
        else:
            dx = to[0] - xy[0]
            dy = to[1] - xy[1]
        for i in range(n):
            y = -i*ndy
            if dy > 0:
                x = dx - i*ndy if dx < 0 else dx + i*ndy
            else:
                x = dx + (n-i-1)*ndy if dx > 0 else dx - (n-i-1)*ndy
            self.segments.append(Segment([(0, y), (x, y), (x, dy)], arrow=arrow))

        self.anchors['mid'] = (dx/2, 0)
        self.elmparams['lblloc'] = 'mid'
        self.elmparams['drop'] = (x, dy)
        self.elmparams['droptheta'] = 90 if dy > 0 else -90
        return super()._place(dwgxy, dwgtheta, **dwgparams)


class Header(Element):
    ''' Header connector element

        Args:
            rows: Number of rows
            cols: Number of columns. Pin numbering requires 1 or 2 columns
            pinsleft: List of pin labels for left side
            pinsright: List of pin labels for right side
            style: Connector style, 'round', 'square', or 'screw' [default: round]
            numbering: Pin numbering order. 'lr' for left-to-right numbering,
                'ud' for up-down numbering, or 'ccw' for counter-clockwise
                (integrated-circuit style) numbering. Pin 1 is always at the
                top-left corner, unless `flip` method is also called. [default: lr]
            shownumber: Draw pin numbers outside the header
            pinalignleft: Vertical alignment for pins on left side
                ('center', 'top', 'bottom') [default: 'bottom']
            pinalignright: Vertical alignment for pins on right side
                ('center', 'top', 'bottom') [default: 'bottom']
            pinfontsizeleft: Font size for pin labels on left [default: 9]
            pinfontsizeright: Font size for pin labels on right [default: 9]
            pinspacing: Distance between pins [default: 0.6]
            edge: Distance between header edge and first pin row/column [default: 0.3]
            pinfill: Color to fill pin circles [default: bg]

        Anchors:
            pin[X] for each pin
    '''
    _element_defaults = {
        'pinspacing': 0.6,
        'style': 'round',
        'numbering': 'lr',
        'pinfontsizeleft': 9,
        'pinfontsizeright': 9,
        'edge': 0.3,
        'pinfill': 'bg',
        'shownumber': False,
        'pinalignleft': 'bottom',
        'pinalignright': 'bottom',
        'pinrad': 0.1
    }
    def __init__(self,
                 rows: int = 4,
                 cols: int = 1,
                 pinsleft: Optional[Sequence[str]] = None,
                 pinsright: Optional[Sequence[str]] = None,
                 *,
                 style: Optional[HeaderStyle] = None,
                 numbering: Optional[HeaderNumbering] = None,
                 shownumber: Optional[bool] = None,
                 pinalignleft: Optional[Valign] = None,
                 pinalignright: Optional[Valign] = None,
                 pinfontsizeright: Optional[float] = None,
                 pinfontsizeleft: Optional[float] = None,
                 pinspacing: Optional[float] = None,
                 edge: Optional[float] = None,
                 pinfill: Optional[str] = None,
                 **kwargs):
        super().__init__(**kwargs)
        if pinsleft is None:
            pinsleft = []
        if pinsright is None:
            pinsright = []
        if cols > 2:
            warnings.warn('Header numbering not supported with cols > 2')

        self.elmparams['d'] = 'right'
        w = (cols-1) * self.params['pinspacing'] + self.params['edge']*2
        h = (rows-1) * self.params['pinspacing'] + self.params['edge']*2
        pinrad = self.params['pinrad']

        self.segments.append(SegmentPoly([(0, 0), (0, h), (w, h), (w, 0)]))
        for row in range(rows):
            for col in range(cols):
                xy = (col*self.params['pinspacing']+self.params['edge'],
                      h-row*self.params['pinspacing']-self.params['edge'])

                if self.params['style'] == 'square':
                    x, y = xy
                    self.segments.append(SegmentPoly([(x-pinrad, y-pinrad), (x+pinrad, y-pinrad),
                                                      (x+pinrad, y+pinrad), (x-pinrad, y+pinrad)],
                                                     fill='bg', zorder=4))
                elif self.params['style'] == 'screw':
                    x, y = xy
                    self.segments.append(SegmentCircle(xy, pinrad*1.75, fill=self.params['pinfill'], zorder=4))
                    self.segments.append(Segment([(x+pinrad, y+pinrad), (x-pinrad, y-pinrad)], zorder=5))

                else:  # self.params['style'] == 'round'
                    self.segments.append(SegmentCircle(xy, pinrad, fill=self.params['pinfill'], zorder=4))

                if self.params['numbering'] == 'lr' or self.params['numbering'] is True:
                    pnumber = str(row*cols + col + 1)
                elif self.params['numbering'] == 'ud':
                    pnumber = str(row + col*rows + 1)
                else:  # number == 'ccw'
                    pnumber = str(rows*col+(rows-row)) if col % 2 else str(row+1)
                self.anchors['pin{}'.format(pnumber)] = xy

                if self.params['shownumber']:
                    numxy = (w+.05 if col % 2 else -.05, xy[1])
                    align = ('left' if col % 2 else 'right', 'bottom')
                    self.segments.append(SegmentText(
                        numxy, pnumber, fontsize=self.params['pinfontsizeleft'], align=align))  # type: ignore

                if pinsleft and (cols == 1 or not col % 2):
                    lblxy = (-.05, xy[1])
                    self.segments.append(SegmentText(lblxy, pinsleft[row], fontsize=self.params['pinfontsizeleft'],
                                                     align=('right', self.params['pinalignleft'])))

                if pinsright and (cols == 1 or col % 2):
                    lblxy = (w+.05, xy[1])
                    self.segments.append(SegmentText(lblxy, pinsright[row], fontsize=self.params['pinfontsizeright'],
                                                     align=('left', self.params['pinalignleft'])))


class Jumper(Element):
    ''' Jumper for use on a Header element

        Set position using `at()` method with a Header
        pin location, e.g. `Jumper().at(H.in1)`

        Args:
            pinspacing: Spacing between pins [default: 0.6]
    '''
    _element_defaults = {
        'pinspacing': 0.6,
        'pinrad': 0.1,
        'theta': 0
    }
    def __init__(self, *,
                 pinspacing: Optional[float] = None,
                 **kwargs):
        super().__init__(**kwargs)
        pinrad = self.params['pinrad']
        x = pinrad*2.5
        self.segments.append(SegmentPoly([(-x, -x), (self.params['pinspacing']+x, -x),
                                         (self.params['pinspacing']+x, x), (-x, x)]))


class BusConnect(Element):
    ''' Data bus connection.

        Adds the short diagonal lines that break out a bus (wide line)
        to connect to an Ic or Header element.

        Args:
            n: Number of parallel lines
            up: Slant up or down
            dy: Distance between parallel lines [default: 0.6]
            lwbus: Line width of bus line [default: 4]
            l: length of connection lines [default: 3]

        Anchors:
            * start
            * end
            * p[X] where X is int for each data line
    '''
    _element_defaults = {
        'dy': 0.6,
        'lwbus': 4,
        'l': 3,
        'slantx': 0.5
    }
    def __init__(self,
                 n: int = 1,
                 up: bool = True,
                 *,
                 dy: Optional[float] = None,
                 lwbus: Optional[float] = None,
                 l: Optional[float] = None,
                 **kwargs):
        super().__init__(**kwargs)
        self.params['theta'] = 0

        deltax = self.params['l']
        deltay = self.params['dy']
        lw = self.params['lwbus']
        slantx = self.params['slantx']
        slanty = slantx if up else -slantx

        for i in range(n):
            y = -i*deltay
            self.segments.append(Segment([(0, y), (deltax-slantx, y), (deltax, y+slanty)]))
            self.anchors[f'pin{i+1}'] = (0, y)
        self.segments.append(Segment([(deltax, slantx), (deltax, slanty-n*deltay)], lw=lw))
        self.elmparams['drop'] = (deltax, slantx)
        self.anchors['start'] = (deltax, slantx)
        self.anchors['end'] = (deltax, slantx-n*deltay)


class BusLine(Line):
    ''' Data bus line. Just a wide line.

        Use BusConnect to break out connections to the BusLine.

        Args:
            lw: Line width [default: 4]
    '''
    _element_defaults = {
        'lw': 4
    }
    def __init__(self, lw: Optional[float] = None, **kwargs):
        super().__init__(**kwargs)


class DB9(Element):
    ''' DB9 Connector

        Args:
            pinspacing: Distance between pins [default: 0.6]
            edge: Distance between edge and pins [default: 0.3]
            number: Draw pin numbers [default: False]
            pinfill: Color to fill pin circles [default: bg]

        Anchors:
            * pin1 thru pin9
    '''
    _element_defaults = {
        'pinspacing': 0.6,
        'edge': 0.3,
        'pinfill': 'bg',
        'pinrad': 0.1,
        'number': False,
        'theta': 0,
    }
    def __init__(self, *,
                 pinspacing: Optional[float] = None,
                 edge: Optional[float] = None,
                 number: Optional[bool] = None,
                 pinfill: Optional[str] = None,
                 **kwargs):
        super().__init__(**kwargs)
        spacing = self.params['pinspacing']
        edgepad = self.params['edge']
        fill = self.params['pinfill']
        w = spacing + edgepad*2
        h1 = 4 * spacing + edgepad*2
        h2 = h1 + .5
        pinrad = self.params['pinrad']

        self.segments.append(SegmentPoly([(0, 0), (0, h1), (w, h2), (w, -.5)], cornerradius=.25))

        for i in range(4):
            xy = (edgepad, h1-(i+.5)*spacing-edgepad)
            self.segments.append(SegmentCircle(xy, pinrad, fill=fill, zorder=4))
            self.anchors[f'pin{9-i}'] = xy
            if self.params['number']:
                self.segments.append(SegmentText((xy[0], xy[1]+pinrad),
                                                 str(9-i), fontsize=9,
                                                 align=('center', 'bottom')))
        for i in range(5):
            xy = (edgepad+spacing, h2-(i+.75)*spacing-edgepad)
            self.segments.append(SegmentCircle(xy, pinrad, fill=fill, zorder=4))
            self.anchors[f'pin{5-i}'] = xy
            if self.params['number']:
                self.segments.append(SegmentText((xy[0], xy[1]+pinrad),
                                                 str(5-i), fontsize=9,
                                                 align=('center', 'bottom')))


class DB25(Element):
    ''' DB25 Connector

        Args:
            pinspacing: Distance between pins [default: 0.6]
            edge: Distance between edge and pins [default: 0.3]
            number: Draw pin numbers
            pinfill: Color to fill pin circles [default: bg]

        Anchors:
            * pin1 thru pin25
    '''
    _element_defaults = {
        'pinspacing': 0.6,
        'edge': 0.3,
        'pinfill': 'bg',
        'number': False,
        'pinrad': 0.1,
        'theta': 0
    }
    def __init__(self, *,
                 pinspacing: Optional[float] = None,
                 edge: Optional[float] = None,
                 number: Optional[bool] = None,
                 pinfill: Optional[str] = None,
                 **kwargs):
        super().__init__(**kwargs)
        spacing = self.params['pinspacing']
        edgepad = self.params['edge']
        fill = self.params['pinfill']
        w = spacing + edgepad*2
        h1 = 12 * spacing + edgepad*2
        h2 = h1 + .5
        pinrad = self.params['pinrad']

        self.segments.append(SegmentPoly([(0, 0), (0, h1), (w, h2), (w, -.5)], cornerradius=.25))

        for i in range(12):
            xy = (edgepad, h1-(i+.5)*spacing-edgepad)
            self.segments.append(SegmentCircle(xy, pinrad, fill=fill, zorder=4))
            self.anchors[f'pin{25-i}'] = xy
            if number:
                self.segments.append(SegmentText((xy[0], xy[1]+pinrad),
                                                 str(25-i), fontsize=9,
                                                 align=('center', 'bottom')))
        for i in range(13):
            xy = (edgepad+spacing, h2-(i+.75)*spacing-edgepad)
            self.segments.append(SegmentCircle(xy, pinrad, fill=fill, zorder=4))
            self.anchors[f'pin{13-i}'] = xy
            if number:
                self.segments.append(SegmentText((xy[0], xy[1]+pinrad),
                                                 str(13-i), fontsize=9,
                                                 align=('center', 'bottom')))


class CoaxConnect(Element):
    ''' Coaxial connector

        Args:
            radius: Radius of outer shell [defualt: 0.4]
            radiusinner: Radius of inner conductor [default: 0.12]
            fillinner: Color to fill inner conductor [default: bg]

        Anchors:
            * center
            * N
            * S
            * E
            * W
    '''
    _element_defaults = {
        'radius': 0.4,
        'radiusinner': 0.12,
        'fillinner': 'bg'
    }
    def __init__(self, *,
                 radius: Optional[float] = None, 
                 radiusinner: Optional[float] = None,
                 fillinner: Optional[str] = None,
                 **kwargs):
        super().__init__(**kwargs)
        r1 = self.params['radius']
        r2 = self.params['radiusinner']
        fill = self.params['fillinner']
        self.segments.append(SegmentCircle((0, 0), r1))
        self.segments.append(SegmentCircle((0, 0), r2, fill=fill, zorder=4))
        self.anchors['center'] = (0, 0)
        self.anchors['N'] = (0, r1)
        self.anchors['S'] = (0, -r1)
        self.anchors['E'] = (r1, 0)
        self.anchors['W'] = (-r1, 0)


class Plug(Element):
    ''' Plug (male connector) '''
    _element_defaults = {
        'pluggap': 0.18
    }
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        pluggap = self.params['pluggap']
        self.segments.append(Segment([(0, 0), (1, 0), gap, (1-resheight, resheight), (1, 0),
                                      (1-resheight, -resheight)]))
        self.elmparams['drop'] = (1+pluggap, 0)


class Jack(Element):
    ''' Jack (female connector) '''
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.segments.append(Segment([(0, 0), (-resheight, resheight), gap,
                                      (-resheight, -resheight), (0, 0), (1, 0)]))
        self.elmparams['drop'] = (1, 0)


class Terminal(Element):
    ''' Terminal element
    
        Args:
            r: Radius of terminal [default: 0.18]
            open: Draw as open circle
    '''
    _element_defaults = {
        'r': 0.18,
        'open': True,
        'fill': 'bg',
        'drop': (0, 0),
        'theta': 0,
        'zorder': 4,
    }
    def __init__(self, *,
                 r: Optional[float] = None,
                 open: Optional[bool] = None,
                 **kwargs):
        super().__init__(**kwargs)
        radius = self.params['r']
        self.anchors['start'] = (0, 0)
        self.anchors['center'] = (0, 0)
        self.anchors['end'] = (0, 0)
        self.elmparams['fill'] = 'bg' if self.params['open'] else True
        self.segments.append(SegmentCircle((0, 0), radius))
        kx = 1.0
        ky = 2.0
        self.segments.append(Segment(
            [(-kx*radius, -ky*radius),
             ( kx*radius,  ky*radius)]))


''' Schemdraw transformations for converting local element definition to
    global position within the drawing
'''

from __future__ import annotations
from typing import Sequence

from .util import Point
from .types import XY


class Transform:
    ''' Class defining transformation matrix

        Args:
            theta: Rotation angle in degrees
            globalshift: X-Y shift (applied after zoom and rotation)
            localshift: Local X-Y shift (applied before zoom and rotation)
            zoom: Zoom factor
    '''
    def __init__(self, theta: float, globalshift: XY,
                 localshift: XY = (0, 0), zoom: XY | float = Point((1, 1))):
        self.theta = theta
        self.shift = Point(globalshift)
        self.localshift = Point(localshift)
        if isinstance(zoom, (int, float)):
            zoom = Point((zoom, zoom))
        self.zoom = zoom

    def __repr__(self):
        return f'Transform: xy={self.shift}; theta={self.theta}; scale={self.zoom}; lshift={self.localshift}'

    def transform(self, pt: XY) -> Point:
        ''' Apply the transform to the point

            Args:
            pt: Original (x, y) coordinates

            Returns:
                Transformed (x, y) coordinates
        '''
        return ((Point(pt) + self.localshift) * self.zoom).rotate(self.theta) + self.shift

    def transform_array(self, pts: Sequence[XY]) -> list[Point]:
        ''' Apply the transform to multiple points

            Args:
                pts: List of (x,y) points to transform

            Returns:
                List of transformed (x, y) points
        '''
        return [self.transform(pt) for pt in pts]



''' SVG drawing backend for schemdraw '''

from __future__ import annotations

from typing import Sequence, Optional, BinaryIO
from xml.etree import ElementTree as ET
from collections import namedtuple

import os
import sys
import subprocess
import tempfile
import math
import warnings
import base64

try:
    import ziamath  # type: ignore
except ImportError:
    ziamath = None  # type: ignore

from ..types import Capstyle, Joinstyle, Linestyle, BBox, Halign, Valign, RotationMode, TextMode, XY
from ..util import Point
from . import svgtext
from .svgunits import parse_size_to_px, PT_PER_IN, PX_PER_PT


LINE_WIDTH = 2     # Default line width is 2 points


class Config:
    ''' Configuration options for SVG backend '''
    _text: TextMode = 'path' if ziamath is not None else 'text'

    @property
    def text(self) -> TextMode:
        ''' One of 'path' or 'text'. In 'text' mode, text is drawn
            as SVG <text> elements and will be searchable in the
            SVG, however it may render differently on systems without
            the same fonts installed. In 'path' mode, text is
            converted to SVG <path> elements and will render
            independently of any fonts on the system. Path mode
            enables full rendering of math expressions, but also
            requires the ziafont/ziamath packages.
        '''
        return self._text

    @text.setter
    def text(self, value: TextMode) -> None:
        if value == 'path' and ziamath is None:
            raise ValueError('Path mode requires ziamath package')
        if value not in ['path', 'text']:
            raise ValueError('text mode must be "path" or "text".')
        self._text = value

    @property
    def svg2(self) -> bool:
        ''' Use SVG2.0. Disable for better browser compatibility
            at the expense of SVG size.
        '''
        if ziamath is not None:
            return ziamath.config.svg2
        return True

    @svg2.setter
    def svg2(self, value: bool) -> None:
        if ziamath is None:
            raise ValueError('SVG2 mode requires ziamath package')
        ziamath.config.svg2 = value

    @property
    def precision(self) -> float:
        ''' Decimal precision for SVG coordinates '''
        return ziamath.config.precision

    @precision.setter
    def precision(self, value: float) -> None:
        ziamath.config.precision = value


config = Config()


hatchpattern = '''<defs><pattern id="hatch" patternUnits="userSpaceOnUse" width="4" height="4">
<path d="M-1,1 l2,-2 M0,4 l4,-4 M3,5 l2,-2" style="stroke:black; stroke-width:.5" /></pattern></defs>'''


def settextmode(mode: TextMode, svg2: bool = True) -> None:
    ''' Set the mode for rendering text in the SVG backend.

        In 'text' mode, text is drawn as SVG <text> elements
        and will be searchable in the SVG, however it may
        render differently on systems without the same fonts
        installed. In 'path' mode, text converted to SVG
        <path> elements and will render independently of
        any fonts on the system. Path mode enables full
        rendering of math expressions, but also requires the
        ziafont/ziamath packages.

        Args:
            mode: Text Mode.
            svg2: Use SVG2.0. Disable for better compatibility.
    '''
    warnings.warn('settextmode is deprecated. Use schemdraw.svgconfig',  DeprecationWarning)
    config.svg2 = svg2
    config.text = mode


def isnotebook():
    ''' Determine whether code is running in Jupyter/interactive mode '''
    try:
        shell = get_ipython().__class__.__name__
        return shell in ['ZMQInteractiveShell', 'SpyderShell']
    except NameError:
        return False


inline = isnotebook()


def getstyle(color: Optional[str] = None, ls: Optional[Linestyle] = None, lw: Optional[float] = None,
             capstyle: Optional[Capstyle] = None, joinstyle: Optional[Joinstyle] = None,
             fill: Optional[str] = None, hatch: bool = False) -> str:
    ''' Get style for svg element. Leave empty if property matches default '''
    # Note: styles are added to every SVG element, rather than in a global <style>
    # tag, since multiple images in one HTML page may share <styles>.
    s = ''
    if isinstance(color, tuple):
        s += f'stroke:rgb({int(color[0]*255)},{int(color[1]*255)},{int(color[2]*255)});'  
    elif color:
        s += f'stroke:{color};'
    if hatch:
        s += 'fill:url(#hatch);'
    else:
        s += f'fill:{str(fill).lower()};'
    if lw:
        s += f'stroke-width:{lw};'
    dash = {'--': '7.4,3.2',
            'dashed': '7.4,3.2',
            ':': '2,3.3',
            'dotted': '2,3.3',
            '-.': '12.8,3.2,2,3.2',
            'dashdot': '12.8,3.2,2,3.2',
            }.get(ls, None)  # type: ignore
    if dash is not None:
        s += f'stroke-dasharray:{dash};'
    if capstyle:
        if capstyle == 'projecting':  # Matplotlib notation
            capstyle = 'square'
        s += f'stroke-linecap:{capstyle};'
    if joinstyle:
        s += f'stroke-linejoin:{joinstyle};'

    return s


def text_size(text: str,
              font: Optional[str] = 'sans',
              mathfont: Optional[str] = None,
              size: float = 14) -> tuple[float, float, float]:
    ''' Get size of text. Size will be exact bounding box if ziamath installed and
        using path text mode. Otherwise size will be estimated based on character
        widths.

        Args:
            text: string to calculate
            font: Font family
            size: Font size in points

        Returns:
            width, height, descender
    '''
    if font is None or font.lower() in ['sans-serif', 'Arial']:
        font = 'sans'

    if (ziamath and
        (mathfont is None or os.path.exists(mathfont))):
        if text == '':
            return (0, 0, 0)

        m = ziamath.Text(text, size=size, mathstyle=font, textfont=font, mathfont=mathfont)
        return (*m.getsize(), m.getyofst())

    return svgtext.text_approx_size(text, font=font, size=size)


class Figure:
    ''' Schemdraw figure drawn directly to SVG

        Args:
            bbox: Coordinate bounding box for drawing
            inches_per_unit: Scale for the drawing
            showbbox: Show frame around entire drawing
    '''
    # Keep track of clipid's across all figures so they don't conflict
    # when multiple figures are in one Jupyter notebook/html file.
    total_clips = 0

    def __init__(self, bbox: BBox, **kwargs):
        self.svgelements: list[tuple[int, ET.Element]] = []  # (zorder, element)
        self.hatch: bool = False
        self.clips: dict[BBox, int] = {}
        self.showbbox = kwargs.get('showbbox', False)
        self.scale = PT_PER_IN * kwargs.get('inches_per_unit', 0.5)   # Converts drawing units to points
        self.margin = kwargs.get('margin', 0.1) + LINE_WIDTH/PT_PER_IN  # Margin in drawing units. Add line width (2pt) to include linecaps in bbox
        self.set_bbox(bbox)
        self._bgcolor: Optional[str] = None
        self._need_xlink = False
        self.svgcanvas = kwargs.get('svg')

    def set_bbox(self, bbox: BBox) -> None:
        ''' Set the bounding box '''
        self.bbox = bbox
        self.bbox = BBox(self.bbox.xmin - self.margin,
                         self.bbox.ymin - self.margin,
                         self.bbox.xmax + self.margin,
                         self.bbox.ymax + self.margin)

        self.pxwidth = (self.bbox.xmax-self.bbox.xmin) * self.scale
        self.pxheight = (self.bbox.ymax-self.bbox.ymin) * self.scale
        self.pxwidth = max(5, self.pxwidth)
        self.pxheight = max(5, self.pxheight)

    def xform(self, x: float, y: float) -> Point:
        ''' Convert x, y in user coords to svg pixel coords '''
        return Point((x*self.scale, -y*self.scale))

    def bgcolor(self, color: str) -> None:
        ''' Set background color of drawing '''
        self._bgcolor = color

    def addclip(self, et: ET.Element, bbox: Optional[BBox]):
        ''' Add clip path to the element '''
        if bbox is not None:
            if bbox in self.clips:
                clipid = self.clips[bbox]
            else:
                clipid = Figure.total_clips
                Figure.total_clips += 1
                self.clips[bbox] = clipid

                x0, y0 = self.xform(bbox.xmin, bbox.ymin)
                x1, y1 = self.xform(bbox.xmax, bbox.ymax)
                clip = ET.fromstring(f'''<defs><clipPath id="clip{clipid}"><rect x="{x0-1}" y="{y0-1}"'''
                                     f''' width="{x1-x0+2}" height="{y1-y0+2}" /></clipPath></defs>''')
                self.svgelements.append((0, clip))
            et.set('clip-path', f'url(#clip{clipid})')

    def plot(self, x: XY, y: XY,
             color: str = 'black', ls: Linestyle = '-', lw: float = 2,
             fill: str = 'none', capstyle: Capstyle = 'round',
             joinstyle: Joinstyle = 'round', clip: Optional[BBox] = None, zorder: int = 2) -> None:
        ''' Plot a path '''
        et = ET.Element('path')
        d = 'M {},{} '.format(*self.xform(x[0], y[0]))
        for xx, yy in zip(x[1:], y[1:]):
            if str(xx) == 'nan' or str(yy) == 'nan':
                d += 'M '
                continue
            elif not d.endswith('M '):
                d += 'L '
            xx, yy = self.xform(xx, yy)
            d += f'{xx},{yy} '

        d = d.strip()
        et.set('d', d)
        et.set('style', getstyle(color=color, ls=ls, lw=lw, capstyle=capstyle,
                                 joinstyle=joinstyle, fill=fill))
        self.addclip(et, clip)
        self.svgelements.append((zorder, et))

    def text(self, s: str, x: float, y: float, color: str = 'black',
             fontsize: float = 14, fontfamily: str = 'sans',
             mathfont: Optional[str] = None,
             rotation: float = 0,
             halign: Halign = 'center',
             valign: Valign = 'center',
             rotation_mode: RotationMode = 'anchor',
             clip: Optional[BBox] = None, zorder: int = 3) -> None:
        ''' Add text to the figure '''
        if s == '':
            return

        x0, y0 = self.xform(x, y)
        if fontfamily.lower() in ['sans-serif', 'Arial']:
            fontfamily = 'sans'

        if valign in ['base', 'baseline']:
            # Matplotlib's "base" valign aligns to the bottom row
            # Ziafont valign aligns to top row.. shift y to compensate and
            # match MPL.
            nlines = len(s.splitlines())
            y0 -= (nlines-1)*fontsize

        if ziamath and config.text == 'path':
            texttag = ET.Element('g')
            ztext = ziamath.Text(s, textfont=fontfamily, mathfont=mathfont,
                                 size=fontsize, linespacing=1, color=color,
                                 rotation=rotation, rotation_mode=rotation_mode)
            ztext.drawon(texttag, x0, y0,
                         halign=halign, valign=valign)
        else:
            texttag = svgtext.text_tosvg(s, x0, y0, font=fontfamily, size=fontsize,
                                         halign=halign, valign=valign, color=color,
                                         rotation=rotation, rotation_mode=rotation_mode,
                                         testmode=False)
        
        self.addclip(texttag, clip)
        self.svgelements.append((zorder, texttag))

    def poly(self, verts: Sequence[XY], closed: bool = True,
             color: str = 'black', fill: str = 'none', lw: float = 2,
             ls: Linestyle = '-', hatch: bool = False, capstyle: Capstyle = 'round',
             joinstyle: Joinstyle = 'round', clip: Optional[BBox] = None, zorder: int = 1) -> None:
        ''' Draw a polygon '''
        et = ET.Element('polyline') if not closed else ET.Element('polygon')
        points = ''
        for xx, yy in verts:
            xx, yy = self.xform(xx, yy)
            points += f'{xx},{yy} '
        et.set('points', points)
        et.set('style', getstyle(color=color, ls=ls, lw=lw, capstyle=capstyle,
                                 joinstyle=joinstyle, fill=fill, hatch=hatch))
        self.addclip(et, clip)
        self.svgelements.append((zorder, et))
        if hatch:
            self.hatch = True

    def circle(self, center: XY, radius: float, color: str = 'black',
               fill: str = 'none', lw: float = 2, ls: Linestyle = '-',
               clip: Optional[BBox] = None, zorder: int = 1) -> None:
        ''' Draw a circle '''
        x, y = self.xform(*center)
        radius = radius * self.scale
        et = ET.Element('circle')
        et.set('cx', str(x))
        et.set('cy', str(y))
        et.set('r', str(radius))
        et.set('style', getstyle(color=color, lw=lw, ls=ls, fill=fill))
        self.addclip(et, clip)
        self.svgelements.append((zorder, et))

    def arrow(self, xy: XY, theta: float,
              arrowwidth: float = .15, arrowlength: float = .25,
              color: str = 'black', lw: float = 2, clip: Optional[BBox] = None, zorder: int = 1) -> None:
        ''' Draw an arrowhead '''
        x, y = self.xform(*xy)
        dx = arrowlength/2 * math.cos(math.radians(theta)) * self.scale
        dy = arrowlength/2 * math.sin(math.radians(theta)) * self.scale
        arrowwidth = arrowwidth*self.scale
        arrowlength = arrowlength*self.scale

        # Draw arrow as path
        head = Point((x, y))
        tail = Point((x-dx, y+dy))
        fullen = math.sqrt(dx**2 + dy**2)
        theta = -math.degrees(math.atan2(dy, dx))

        fin1 = Point((fullen - arrowlength, arrowwidth/2)).rotate(theta) + tail
        fin2 = Point((fullen - arrowlength, -arrowwidth/2)).rotate(theta) + tail

        # Shrink arrow head by lw so it points right at the line
        head = Point((head[0] - lw * 2 * math.cos(math.radians(theta)),
                      head[1] - lw * 2 * math.sin(math.radians(theta))))

        et1 = ET.Element('path')
        d = f'M {head[0]} {head[1]} '
        d += f'L {fin1[0]} {fin1[1]} '
        d += f'L {fin2[0]} {fin2[1]} Z'
        et1.set('d', d)
        et1.set('style', getstyle(color=color, lw=0, capstyle='butt',
                                  joinstyle='miter', fill=color))
        self.addclip(et1, clip)
        self.svgelements.append((zorder, et1))

    def bezier(self, p: Sequence[Point], color: str = 'black',
               lw: float = 2, ls: Linestyle = '-', capstyle: Capstyle = 'round', zorder: int = 1,
               arrow: Optional[str] = None, arrowlength=.25, arrowwidth=.15, clip: Optional[BBox] = None) -> None:
        ''' Draw a cubic or quadratic bezier '''
        # Keep original points for arrow head
        # and adjust points for line so they don't extrude from arrows.
        lpoints = list(p)
        if arrow is not None:
            if '<' in arrow:
                th1 = math.atan2(p[0].y - p[1].y, p[0].x - p[1].x)
                lpoints[0] = Point((p[0].x - math.cos(th1) * arrowlength/2,
                                    p[0].y - math.sin(th1) * arrowlength/2))
            if '>' in arrow:
                th2 = math.atan2(p[-1].y - p[-2].y, p[-1].x - p[-2].x)
                lpoints[-1] = Point((p[-1].x - math.cos(th2) * arrowlength/2,
                                    p[-1].y - math.sin(th2) * arrowlength/2))

        lpoints = [self.xform(*p0) for p0 in lpoints]
        order = 'C' if len(p) == 4 else 'Q'

        et = ET.Element('path')
        path = f'M {lpoints[0][0]} {lpoints[0][1]} {order}'
        for p0 in lpoints[1:]:
            path += f' {p0[0]} {p0[1]}'
        et.set('d', path)
        et.set('style', getstyle(color=color, ls=ls, lw=lw, capstyle=capstyle))
        self.addclip(et, clip)
        self.svgelements.append((zorder, et))

        if arrow is not None:
            # Note: using untransformed bezier control points here
            if '<' in arrow:
                delta = p[0] - p[1]
                theta = math.degrees(math.atan2(delta.y, delta.x))
                self.arrow(p[0], theta, color=color, lw=1, zorder=zorder,
                           clip=clip, arrowlength=arrowlength, arrowwidth=arrowwidth)
            elif arrow.startswith('o'):
                self.circle(p[0], radius=arrowwidth/2, color=color, fill=color, lw=0,
                            clip=clip, zorder=zorder)
            if '>' in arrow:
                delta = p[-1] - p[-2]
                theta = math.degrees(math.atan2(delta.y, delta.x))
                self.arrow(p[-1], theta, color=color, lw=1, zorder=zorder,
                           clip=clip, arrowlength=arrowlength, arrowwidth=arrowwidth)
            elif arrow.endswith('o'):
                self.circle(p[-1], radius=arrowwidth/2, color=color, fill=color, lw=0,
                            clip=clip, zorder=zorder)

    def path(self, path: Sequence[XY | str],
            color: str = 'black', lw: float = 2, ls: Linestyle = '-',
            fill: Optional[str] = None,
            capstyle: Capstyle = 'round',
            joinstyle: Joinstyle = 'round', 
            zorder: int = 1,
            clip: Optional[BBox] = None,
            ) -> None:
        dstrs = []
        for point in path:
            if isinstance(point, str):
                dstrs.append(point)
            else:
                x, y = self.xform(*point)
                dstrs.append(f'{x}')
                dstrs.append(f'{y}')
        et = ET.Element('path')
        et.set('d', ' '.join(dstrs))
        et.set('style', getstyle(color=color, ls=ls, lw=lw, capstyle=capstyle,
                                 joinstyle=joinstyle, fill=fill))
        self.addclip(et, clip)
        self.svgelements.append((zorder, et))

    def arc(self, center: XY, width: float, height: float,
            theta1: float = 0, theta2: float = 90, angle: float = 0,
            color: str = 'black', lw: float = 2, ls: Linestyle = '-',
            fill: Optional[str] = None,
            zorder: int = 1,
            arrow: Optional[str] = None, clip: Optional[BBox] = None) -> None:
        ''' Draw an arc or ellipse, with optional arrowhead '''
        centerx, centery = self.xform(*center)
        width, height = width*self.scale, height*self.scale
        angle = -angle
        theta1 = -theta1
        theta2 = -theta2
        anglerad = math.radians(angle)

        theta1 = math.radians(theta1)
        theta2 = math.radians(theta2)
        t1 = math.atan2(width*math.sin(theta1), height*math.cos(theta1))
        t2 = math.atan2(width*math.sin(theta2), height*math.cos(theta2))
        while t1 < t2:
            t1 += 2*math.pi

        startx = (centerx + width/2 * math.cos(t2)*math.cos(anglerad)
                  - height/2 * math.sin(t2)*math.sin(anglerad))
        starty = (centery + width/2 * math.cos(t2)*math.sin(anglerad)
                  + height/2 * math.sin(t2)*math.cos(anglerad))
        endx = (centerx + width/2 * math.cos(t1)*math.cos(anglerad)
                - height/2 * math.sin(t1)*math.sin(anglerad))
        endy = (centery + width/2 * math.cos(t1)*math.sin(anglerad)
                + height/2 * math.sin(t1)*math.cos(anglerad))

        startx, starty = round(startx, 2), round(starty, 2)
        endx, endy = round(endx, 2), round(endy, 2)
        dx, dy = endx-startx, endy-starty

        if abs(dx) < .1 and abs(dy) < .1:
            # Full ellipse - can't be drawn with a single <path>
            # because when start/end points are the same it draws a dot.
            et = ET.Element('ellipse')
            et.set('cx', str(centerx))
            et.set('cy', str(centery))
            et.set('rx', str(width/2))
            et.set('ry', str(height/2))
            if angle != 0:
                et.set('transform', f'rotate({angle} {centerx} {centery})')
            et.set('style', getstyle(color=color, ls=ls, lw=lw, fill=fill))
            self.addclip(et, clip)
            self.svgelements.append((zorder, et))

        else:
            flags = '1 1' if abs(t2-t1) >= math.pi else '0 1'
            et = ET.Element('path')
            d = f'M {startx} {starty}'
            d += f' a {width/2} {height/2} {angle} {flags} {dx} {dy}'
            et.set('d', d)
            et.set('style', getstyle(color=color, ls=ls, lw=lw, fill=fill))
            self.addclip(et, clip)
            self.svgelements.append((zorder, et))

        if arrow is not None:
            # Note: This arrowhead's TAIL is located at the endpoint of the
            # arc curve. The arrow points beyond the arc's theta.
            # Back to user coordinates
            width, height = width/self.scale, height/self.scale
            angle = -angle
            theta1 = -math.degrees(theta1)
            theta2 = -math.degrees(theta2)

            arrowlength = .25
            arrowwidth = .15

            x, y = math.cos(math.radians(theta2)), math.sin(math.radians(theta2))
            th2 = math.degrees(math.atan2((width/height)*y, x))
            x, y = math.cos(math.radians(theta1)), math.sin(math.radians(theta1))
            th1 = math.degrees(math.atan2((width/height)*y, x))
            if arrow in ['ccw', 'end', 'both'] or '>' in arrow:
                dx = math.cos(math.radians(th2+90)) * arrowlength
                dy = math.sin(math.radians(th2+90)) * arrowlength
                xy = Point((center[0] + width/2*math.cos(math.radians(th2)),
                            center[1] + height/2*math.sin(math.radians(th2))))
                xy = Point(xy).rotate(angle, center)
                darrow = Point((dx, dy)).rotate(angle)
                theta = math.degrees(math.atan2(darrow.y, darrow.x))
                self.arrow(xy+darrow, theta, arrowwidth=arrowwidth,
                           arrowlength=arrowlength, color=color, lw=1, zorder=zorder)

            if arrow in ['cw', 'start', 'both'] or '<' in arrow:
                dx = -math.cos(math.radians(th1+90)) * arrowlength
                dy = -math.sin(math.radians(th1+90)) * arrowlength
                xy = Point((center[0] + width/2*math.cos(math.radians(th1)),
                            center[1] + height/2*math.sin(math.radians(th1))))

                xy = Point(xy).rotate(angle, center)
                darrow = Point((dx, dy)).rotate(angle)
                theta = math.degrees(math.atan2(darrow.y, darrow.x))
                self.arrow(xy+darrow, theta, arrowwidth=arrowwidth,
                           arrowlength=arrowlength, color=color, lw=1, zorder=zorder)

    def image(self, image: str | BinaryIO, xy: XY, width: float, height: float,
              rotate: float = 0, zorder: int = 1, imgfmt: Optional[str] = None):
        ''' Add an image to the figure
        
            Args:
                image: File name or open file pointer
                xy: Position for the image
                width: Image width in data coordinates
                height: Image height in data coordinates
                rotate: Image rotation in degrees
                zorder: Z-Order for image
                imgfmt: file format of image. Required if image is
                    open file pointer.
        '''
        if isinstance(image, str):
            with open(image, 'rb') as f:
                imgdat = f.read()
                imgfmt = os.path.splitext(image)[1][1:] if imgfmt is None else imgfmt

        else:
            image.seek(0)
            imgdat = image.read()

        if imgfmt is None:
            raise ValueError('Must specify image format when using file pointer with SegmentImage')

        width = width * self.scale  # Width/height in pixels
        height = height * self.scale
        x0, y0 = self.xform(*xy)
        y0 -= height

        if imgfmt == 'svg':
            et = ET.Element('g')
            imageelm = ET.fromstring(imgdat.decode())
            imgwidth = parse_size_to_px(imageelm.get('width', '0'))
            s = width / imgwidth
            xform = f'translate({x0}, {y0}) scale({s})'
            if rotate:
                xform = f'rotate({-rotate} {x0} {y0+height}) ' + xform
            et.set('transform', xform)
            et.append(imageelm)
        else:  # Raster images
            image_b64 = base64.encodebytes(imgdat).decode()
            et = ET.Element('image')
            et.set('xlink:href', f'data:image/{imgfmt};base64,{image_b64}')
            self.svgelements.append((zorder, et))
            self._need_xlink = True

            et.set('x', str(x0))
            et.set('y', str(y0))
            et.set('width', str(width))
            et.set('height', str(height))
            if rotate:
                et.set('transform', f'rotate({-rotate} {x0} {y0+height})')
        self.svgelements.append((zorder, et))

    def save(self, fname: str, **kwargs) -> None:
        ''' Save the figure to a file '''
        svg = self.getimage().decode()
        ext = os.path.splitext(fname)[1]
        if ext.lower() != '.svg':
            raise ValueError('SVG backend only supports saving SVG format figures.')
        with open(fname, 'w', encoding='utf-8') as f:
            f.write(svg)

    def getsvg(self) -> ET.Element:
        ''' Get the image as SVG XML Tree '''
        x0 = self.bbox.xmin * self.scale
        y0 = -self.bbox.ymax * self.scale
        if not self.svgcanvas:
            svg = ET.Element('svg')
            svg.set('xmlns', 'http://www.w3.org/2000/svg')
            if self._need_xlink:
                svg.set('xmlns:xlink', 'http://www.w3.org/1999/xlink')
            svg.set('xml:lang', 'en')
            svg.set('height', f'{self.pxheight}pt')
            svg.set('width', f'{self.pxwidth}pt')
            svg.set('viewBox', f'{x0} {y0} {self.pxwidth} {self.pxheight}')
            if self._bgcolor:
                svg.set('style', f'background-color:{self._bgcolor};')
        else:
            svg = self.svgcanvas

        if self.hatch:
            svg.append(ET.fromstring(hatchpattern))

        if self.showbbox:
            rect = ET.SubElement(svg, 'rect')
            rect.set('x', str(self.bbox.xmin*self.scale))
            rect.set('y', str(-self.bbox.ymax*self.scale))
            rect.set('width', str(self.bbox.xmax*self.scale - self.bbox.xmin*self.scale))
            rect.set('height', str(-self.bbox.ymin*self.scale + self.bbox.ymax*self.scale))
            rect.set('style', 'fill:none; stroke-width:0.5; stroke:black;')

            rect = ET.SubElement(svg, 'rect')
            rect.set('x', str((self.bbox.xmin+self.margin)*self.scale))
            rect.set('y', str(-(self.bbox.ymax-self.margin)*self.scale))
            rect.set('width', str((self.bbox.xmax-self.margin)*self.scale - (self.bbox.xmin+self.margin)*self.scale))
            rect.set('height', str(-(self.bbox.ymin+self.margin)*self.scale + (self.bbox.ymax-self.margin)*self.scale))
            rect.set('style', 'fill:none; stroke-width:1; stroke:red;')

        # sort by zorder
        elements = [k[1] for k in sorted(self.svgelements, key=lambda x: x[0])]
        for elm in elements:
            svg.append(elm)
        return svg

    def getimage(self, ext: str = 'svg') -> bytes:
        ''' Get image as SVG bytes '''
        if ext.lower() != 'svg':
            raise ValueError('SVG backend only supports generating SVG format figures.')

        svg = self.getsvg()
        return ET.tostring(svg, encoding='utf-8')

    def clear(self) -> None:
        ''' Remove everything '''
        self.svgelements = []

    def _repr_svg_(self):
        ''' SVG representation for Jupyter '''
        return self.getimage().decode()

    def __repr__(self):
        return self.getimage().decode()

    def show(self) -> None:
        ''' Show drawing in default program, if not inline in Jupyter '''
        if not inline:
            handle, path = tempfile.mkstemp(suffix='.svg')
            with os.fdopen(handle, 'w') as f:
                f.write(self.getimage().decode())

            if sys.platform == 'win32':
                os.startfile(path)
            else:
                cmd = 'open' if sys.platform == 'darwin' else 'xdg-open'
                try:
                    subprocess.call([cmd, path])
                except (OSError, FileNotFoundError):  # Some linux without display may not have xdg-open
                    pass



''' Module for converting a logic string expression into a schemdraw.Drawing.

Example:

>>> logicparse("a and (b or c)")

'''
from typing import Optional
import pyparsing  # type: ignore

from .. import schemdraw
from .. import logic
from ..elements import RightLines
from .buchheim import buchheim


class LogicTree():
    ''' Organize the logic gates into tree structure '''
    def __init__(self, node, *children):
        self.node = node
        self.children = children if children else []

    def __getitem__(self, key):
        if isinstance(key, (int, slice)):
            return self.children[key]

    def __iter__(self):
        return self.children.__iter__()

    def __len__(self):
        return len(self.children)


def parse_string(logicstr):
    ''' Parse the logic string using pyparsing '''
    and_ = pyparsing.Keyword('and')
    or_ = pyparsing.Keyword('or')
    nor_ = pyparsing.Keyword('nor')
    nand_ = pyparsing.Keyword('nand')
    xor_ = pyparsing.Keyword('xor')
    xnor_ = pyparsing.Keyword('xnor')
    not_ = pyparsing.Keyword('not')
    true_ = pyparsing.Keyword('true')
    false_ = pyparsing.Keyword('false')

    not_op = not_ | '~' | '¬'
    and_op = and_ | nand_ | '&' | '∧'
    xor_op = xor_ | xnor_ | '⊕' | '⊻'
    or_op = or_ | nor_ | '|' | '∨' | '+'

    expr = pyparsing.Forward()

    identifier = ~(and_ | or_ | nand_ | nor_ | not_ | true_ | false_) + \
        pyparsing.Word('$' + pyparsing.alphas + '_', pyparsing.alphanums + '_' + '$')

    expr = pyparsing.infixNotation(true_ | false_ | identifier,
                                   [(not_op, 1, pyparsing.opAssoc.RIGHT),
                                    (and_op, 2, pyparsing.opAssoc.LEFT),
                                    (or_op, 2, pyparsing.opAssoc.LEFT),
                                    (xor_op, 2, pyparsing.opAssoc.LEFT)])

    return expr.parseString(logicstr)[0]


def to_tree(pres):
    ''' Convert the parsed logic expression into a LogicTree '''
    invertfunc = False

    if pres[0] in ['not', '~', '¬']:
        if isinstance(pres[1], str):
            return LogicTree('not', to_tree(pres[1]))
        else:
            pres = pres[1]
            invertfunc = True

    if isinstance(pres, str):
        return LogicTree(pres)

    func = pres[1]
    inputs = pres[::2]

    func = {'&': 'and', '∧': 'and',
            '|': 'or', '∨': 'or',  '+': 'or',
            '⊕': 'xor', '⊻': 'xor'}.get(func, func)

    if invertfunc:
        func = {'and': 'nand', 'or': 'nor', 'not': 'buf',
                'nand': 'and', 'nor': 'or', 'buf': 'not',
                'xor': 'xnor', 'xnor': 'xor'}.get(func)

    return LogicTree(func, *[to_tree(i) for i in inputs])


def drawlogic(tree, gateH=.7, gateW=2, outlabel=None):
    ''' Draw the LogicTree to a schemdraw Drawing

        Parameters
        ----------
        tree: LogicTree
            The tree structure to draw
        gateH: float
            Height of one gate
        gateW: float
            Width of one gate
        outlabel: string
            Label for logic output

        Returns
        -------
        schemdraw.Drawing
    '''
    drawing = schemdraw.Drawing()
    drawing.unit = gateW  # NOTs still use d.unit

    dtree = buchheim(tree)

    def drawit(root, depth=0, outlabel=None):
        ''' Recursive drawing function '''
        elmdefs = {'and': logic.And,
                   'or': logic.Or,
                   'xor': logic.Xor,
                   'nand': logic.Nand,
                   'xnor': logic.Xnor,
                   'nor': logic.Nor,
                   'not': logic.Not}
        elm = elmdefs.get(root.node, logic.And)

        x = root.y * -gateW   # buchheim draws vertical trees, so flip x-y.
        y = -root.x * gateH

        g = elm(d='r', at=(x, y), anchor='end',
                l=gateW, inputs=len(root.children))
        if outlabel:
            g.label(outlabel, loc='end')
        
        for i, child in enumerate(root.children):
            anchorname = 'start' if elm in [logic.Not, logic.Buf] else f'in{i+1}'
            if child.node not in elmdefs:
                g.label(child.node, loc=anchorname)

        drawing.add(g)

        for i, child in enumerate(root.children):
            anchorname = 'start' if elm in [logic.Not, logic.Buf] else f'in{i+1}'
            if child.node in elmdefs:
                childelm = drawit(child, depth+1)  # recursive
                drawing.add(RightLines(at=(g, anchorname), to=childelm.end))
        return g

    drawit(dtree, outlabel=outlabel)
    return drawing


def logicparse(expr: str, gateW: float = 2, gateH: float = .75,
               outlabel: Optional[str] = None) -> schemdraw.Drawing:
    ''' Parse a logic string expression and draw the gates in a schemdraw Drawing

        Logic expression is defined by string using 'and', 'or', 'not', etc.
        for example, "a or (b and c)". Parser recognizes several symbols and
        names for logic functions:
        [and, '&', '∧']
        [or, '|', '∨', '+']
        [xor, '⊕', '⊻']
        [not, '~', '¬']

        Args:
            expr: Logic expression
            gateH: Height of one gate
            gateW: Width of one gate
            outlabel: Label for logic output

        Returns:
            schemdraw.Drawing with logic tree
    '''
    parsed = parse_string(expr)
    tree = to_tree(parsed)
    drawing = drawlogic(tree, gateH=gateH, gateW=gateW, outlabel=outlabel)
    return drawing



''' Sources, meters, and lamp elements '''

from __future__ import annotations
import math

from .elements import Element2Term, gap
from .twoterm import resheight
from ..segments import Segment, SegmentCircle, SegmentText
from .. import util


class Source(Element2Term):
    ''' Generic source element '''
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.segments.append(Segment([(0, 0), (0, 0), gap, (1, 0), (1, 0)]))
        self.segments.append(SegmentCircle((0.5, 0), 0.5,))
        self.elmparams['theta'] = 90


class SourceV(Source):
    ''' Voltage source '''
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        plus_len = .2
        self.segments.append(Segment([(.25, -plus_len/2),
                                      (.25, plus_len/2)]))    # '-' sign
        self.segments.append(Segment([(.75-plus_len/2, 0),
                                      (.75+plus_len/2, 0)]))  # '+' sign
        self.segments.append(Segment([(.75, -plus_len/2),
                                      (.75, plus_len/2)]))     # '+' sign


class SourceI(Source):
    ''' Current source '''
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.segments.append(Segment([(.25, 0), (.75, 0)], arrow='->'))


class SourceSin(Source):
    ''' Source with sine '''
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        sin_y = util.linspace(-.25, .25, num=25)
        sin_x = [.2 * math.sin((sy-.25)*math.pi*2/.5) + 0.5 for sy in sin_y]
        self.segments.append(Segment(list(zip(sin_x, sin_y))))


class SourcePulse(Source):
    ''' Pulse source '''
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        sq = .15
        x = .4
        self.segments.append(Segment(
            [(x, sq*2), (x, sq), (x+sq, sq), (x+sq, -sq),
             (x, -sq), (x, -sq*2)]))


class SourceTriangle(Source):
    ''' Triangle source '''
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.segments.append(Segment([(.4, .25), (.7, 0), (.4, -.25)]))


class SourceRamp(Source):
    ''' Ramp/sawtooth source '''
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.segments.append(Segment([(.4, .25), (.8, -.2), (.4, -.2)]))


class SourceSquare(Source):
    ''' Square wave source '''
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.segments.append(Segment([(.5, .25), (.7, .25), (.7, 0),
                                      (.3, 0), (.3, -.25), (.5, -.25)]))


class SourceControlled(Element2Term):
    ''' Generic controlled source '''
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.segments.append(Segment([(0, 0), (.5, .5), (1, 0),
                                      (.5, -.5), (0, 0), gap, (1, 0)]))
        self.params['theta'] = 90


class SourceControlledV(SourceControlled):
    ''' Controlled voltage source '''
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        plus_len = .2
        self.segments.append(Segment([(.25, -plus_len/2),
                                      (.25, plus_len/2)]))  # '-' sign
        self.segments.append(Segment([(.75-plus_len/2, 0),
                                      (.75+plus_len/2, 0)]))  # '+' sign
        self.segments.append(Segment([(.75, -plus_len/2),
                                      (.75, plus_len/2)]))  # '+' sign


class SourceControlledI(SourceControlled):
    ''' Controlled current source '''
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.segments.append(Segment([(.25, 0), (.75, 0)], arrow='->'))


batw = resheight*.75
bat1 = resheight*1.5
bat2 = resheight*.75


class BatteryCell(Element2Term):
    ''' Cell '''
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.segments.append(Segment([(0, 0), gap, (batw, 0)]))
        self.segments.append(Segment([(0, bat1), (0, -bat1)]))
        self.segments.append(Segment([(batw, bat2), (batw, -bat2)]))


class Battery(Element2Term):
    ''' Battery '''
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.segments.append(Segment([(0, 0), gap, (batw*3, 0)]))
        self.segments.append(Segment([(0, bat1), (0, -bat1)]))
        self.segments.append(Segment([(batw, bat2), (batw, -bat2)]))
        self.segments.append(Segment([(batw*2, bat1), (batw*2, -bat1)]))
        self.segments.append(Segment([(batw*3, bat2), (batw*3, -bat2)]))


class Solar(Source):
    ''' Solar source '''
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        cellw = resheight*.5
        cellw2 = cellw + .15
        cellx = .4
        self.segments.append(Segment([(cellx, cellw),
                                      (cellx, -cellw)]))
        self.segments.append(Segment([(cellx+.2, cellw2),
                                      (cellx+.2, -cellw2)]))
        self.segments.append(Segment([(0, 0), (cellx, 0), gap,
                                      (cellx+.2, 0), (1, 0)]))
        self.segments.append(Segment([(1.1, .9), (.8, .6)],
                                     arrow='->', arrowwidth=.16, arrowlength=.2))
        self.segments.append(Segment([(1.3, .7), (1, .4)],
                                     arrow='->', arrowwidth=.16, arrowlength=.2))


class MeterV(Source):
    ''' Volt meter '''
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.segments.append(SegmentText((.5, 0), 'V'))


class MeterI(Source):
    ''' Current Meter (I) '''
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.segments.append(SegmentText((.5, 0), 'I'))


class MeterA(Source):
    ''' Ammeter '''
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.segments.append(SegmentText((.5, 0), 'A'))


class MeterOhm(Source):
    ''' Ohm meter '''
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.segments.append(SegmentText((.5, 0), r'$\Omega$'))


class Lamp(Source):
    ''' Incandescent Lamp '''
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        a = .25
        b = .7
        t = util.linspace(1.4, 3.6*math.pi, 100)
        x = [a*t0 - b*math.sin(t0) for t0 in t]
        y = [a - b * math.cos(t0) for t0 in t]
        x = [xx - x[0] for xx in x]  # Scale to about the right size
        x = [xx / x[-1] for xx in x]
        y = [(yy - y[0]) * .25 for yy in y]
        self.segments.append(Segment(list(zip(x, y))))


class Lamp2(Source):
    ''' Incandescent Lamp (with X through a Source) '''
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        r=0.5
        self.segments.append(Segment(
            [(r-r/2**.5, -r/2**.5),
             (r+r/2**.5,  r/2**.5)]))
        self.segments.append(Segment(
            [(r-r/2**.5,  r/2**.5),
             (r+r/2**.5, -r/2**.5)]))


class Neon(Source):
    ''' Neon bulb '''
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        cellw = resheight
        cellx = .4
        self.segments.append(Segment([(cellx, cellw), (cellx, -cellw)]))
        self.segments.append(Segment([(cellx+.2, cellw), (cellx+.2, -cellw)]))
        self.segments.append(Segment([(0, 0), (cellx, 0), gap,
                                      (cellx+.2, 0), (1, 0)]))
        self.segments.append(SegmentCircle((cellx-.15, .2), .05, fill=True))



''' Flowcharting element definitions '''
from __future__ import annotations
from typing import Optional
import math

from ..util import linspace
from ..segments import Segment, SegmentCircle, SegmentPoly, SegmentText
from ..elements import Element
from ..backends import svg


def labelsize(label, pad):
    ''' Get unit size of label and padding '''
    font = 'sans' if label.font is None else label.font
    size = 14 if label.fontsize is None else label.fontsize
    w72, h72, _ = svg.text_size(label.label, font=font, size=size)
    w = w72/64*2 + pad*2
    h = h72/64*2 + pad*2
    return w, h


class Box(Element):
    ''' Flowchart Process Box. Default box has minimum size (3, 2)
        but will expand to fit the label text. Size may be manually
        fixed using w and h arguments.

        Args:
            w: Width of box
            h: Height of box

        Anchors:
            * 16 compass points (N, S, E, W, NE, NNE, etc.)
    '''
    _element_defaults = {
        'w': 3,
        'h': 2,
        'minsize': (3, 2),
        'pad': 0.25,
        'lblloc': 'center',
        'lblofst': 0,
        'theta': 0
    }
    def _set_anchors(self):
        w, h = self.params['w'], self.params['h']
        self.anchors['center'] = (w/2, 0)
        self.anchors['N'] = (w/2, h/2)
        self.anchors['E'] = (w, 0)
        self.anchors['S'] = (w/2, -h/2)
        self.anchors['W'] = (0, 0)
        self.anchors['NW'] = (0, h/2)
        self.anchors['NE'] = (w, h/2)
        self.anchors['SW'] = (0, -h/2)
        self.anchors['SE'] = (w, -h/2)
        self.anchors['NNE'] = (3*w/4, h/2)
        self.anchors['NNW'] = (w/4, h/2)
        self.anchors['SSE'] = (3*w/4, -h/2)
        self.anchors['SSW'] = (w/4, -h/2)
        self.anchors['ENE'] = (w, h/3)
        self.anchors['ESE'] = (w, -h/3)
        self.anchors['WNW'] = (0, h/3)
        self.anchors['WSW'] = (0, -h/3)

    def _set_size(self) -> tuple[float, float]:
        minw, minh = self.params['minsize']
        mainlabels = [label for label in self._userlabels if label.loc is None]
        if mainlabels:
            minw, minh = labelsize(mainlabels[0], self.params['pad'])
            minw = max(minw, self.params['minsize'][0])
            minh = max(minh, self.params['minsize'][1])

        self._userparams.setdefault('w', minw)
        self._userparams.setdefault('h', minh)
        return self._userparams['w'], self._userparams['h']

    def _set_segments(self):
        w, h = self.params['w'], self.params['h']
        self.segments.append(Segment([(0, 0), (0, h/2), (w, h/2),
                                      (w, -h/2), (0, -h/2), (0, 0)]))
    
    def _place(self, dwgxy, dwgtheta, **dwgparams):
        ''' Make the box flow in the current drawing direction '''
        self._set_size()
        self._set_segments()
        self._set_anchors()

        if 'anchor' not in self.params:
            while dwgtheta < 0:
                dwgtheta += 360

            # Pick closest anchor
            thetas = [0, 45, 90, 135, 180, 225, 270, 315]
            anchors = ['W', 'SW', 'S', 'SE', 'E', 'NE', 'N', 'NW']
            idx = min(range(len(thetas)), key=lambda i: abs(thetas[i]-dwgtheta))
            anchor = anchors[idx]
            self.elmparams['anchor'] = anchor
            dropanchor = anchor.translate(anchor.maketrans('NESW', 'SWNE'))
            if dropanchor in self.anchors:
                self.elmparams['drop'] = self.anchors[dropanchor]
        return super()._place(dwgxy, dwgtheta, **dwgparams)


class RoundBox(Box):
    ''' Alternate Process box with rounded corners

        Args:
            w: Width of box
            h: Height of box
            cornerradius: Radius of round corners [default: 0.3]

        Anchors:
            * 16 compass points (N, S, E, W, NE, NNE, etc.)
    '''
    _element_defaults = {
        'cornerradius': 0.3
    }
    def __init__(self, *, cornerradius: Optional[float] = None, **kwargs):
        super().__init__(**kwargs)

    def _set_segments(self):
        w, h = self.params['w'], self.params['h']
        self.segments = [SegmentPoly(
            [(0, h/2), (w, h/2), (w, -h/2), (0, -h/2)],
            cornerradius=self.params['cornerradius'])]
        
    def _set_anchors(self):
        super()._set_anchors()
        w, h = self.params['w'], self.params['h']
        cornerradius = self.params['cornerradius']
        k = cornerradius - cornerradius*math.sqrt(2)/2
        self.anchors['NE'] = (w-k, h/2-k)
        self.anchors['NW'] = (k, h/2-k)
        self.anchors['SE'] = (w-k, -h/2+k)
        self.anchors['SW'] = (k, -h/2+k)
        self.anchors['NNE'] = (3*w/4 if cornerradius < w/4 else w-cornerradius, h/2)
        self.anchors['NNW'] = (w/4 if cornerradius < w/4 else cornerradius, h/2)
        self.anchors['SSE'] = (3*w/4 if cornerradius < w/4 else w-cornerradius, -h/2)
        self.anchors['SSW'] = (w/4 if cornerradius < w/4 else cornerradius, -h/2)
        self.anchors['ENE'] = (w, h/4 if cornerradius < h/4 else h/2-cornerradius)
        self.anchors['ESE'] = (w, -h/4 if cornerradius < h/4 else -h/2+cornerradius)
        self.anchors['WNW'] = (0, h/4 if cornerradius < h/4 else h/2-cornerradius)
        self.anchors['WSW'] = (0, -h/4 if cornerradius < h/4 else -h/2+cornerradius)


class Terminal(RoundBox):
    ''' Flowchart start/end terminal

        Args:
            w: Width of box
            h: Height of box

        Anchors:
            * 16 compass points (N, S, E, W, NE, NNE, etc.)
    '''
    _element_defaults = {
        'w': 3,
        'h': 1.25,
        'minsize': (3, 1.25),
        'droptheta': -90,
        'anchor': 'N'
    }
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def _set_size(self) -> tuple[float, float]:
        w, h = super()._set_size()
        self.elmparams['cornerradius'] = h/2
        return w, h

    def _set_anchors(self):
        super()._set_anchors()
        self.elmparams['drop'] = self.anchors['S']


class Subroutine(Box):
    ''' Flowchart subroutine/predefined process. Box with extra
        vertical lines near sides.

        Args:
            w: Width of box
            h: Height of box
            s: spacing of side lines [default: 0.3]

        Anchors:
            * 16 compass points (N, S, E, W, NE, NNE, etc.)
    '''
    _element_defaults = {
        's': 0.3
    }
    def __init__(self, *, s: Optional[float] = None, **kwargs):
        super().__init__(**kwargs)

    def _set_size(self) -> tuple[float, float]:
        w, h = super()._set_size()
        w += self.params['s']*2
        self._userparams['w'] = w
        return w, h

    def _set_segments(self):
        super()._set_segments()
        w, h = self.params['w'], self.params['h']
        s = self.params['s']
        self.segments.append(Segment([(w-s, h/2), (w-s, -h/2)]))
        self.segments.append(Segment([(s, h/2), (s, -h/2)]))


class Data(Box):
    ''' Flowchart data or input/output box (parallelogram)

        Args:
            w: Width of box
            h: Height of box
            s: slant of sides

        Anchors:
            * 16 compass points (N, S, E, W, NE, NNE, etc.)
    '''
    _element_defaults = {
        'slant': 0.5
    }
    def __init__(self, *, slant: Optional[float] = 0.5, **kwargs):
        super().__init__(**kwargs)

    def _set_segments(self):
        w, h = self.params['w'], self.params['h']
        s = self.params['slant']
        self.segments.append(SegmentPoly([(0, 0), (s/2, h/2), (w+s/2, h/2),
                                          (w-s/2, -h/2), (-s/2, -h/2)]))

    def _set_anchors(self):
        w, h = self.params['w'], self.params['h']
        s = self.params['slant']
        super()._set_anchors()
        self.anchors['N'] = (w/2+s/2, h/2)
        self.anchors['S'] = (w/2-s/2, -h/2)
        self.anchors['NE'] = (w+s/2, h/2)
        self.anchors['SE'] = (w-s/2, -h/2)
        self.anchors['NW'] = (s/2, h/2)
        self.anchors['SW'] = (-s/2, -h/2)
        self.anchors['ENE'] = (w+s/4, h/4)
        self.anchors['WNW'] = (s/4, h/4)
        self.anchors['ESE'] = (w-s/4, -h/4)
        self.anchors['WSW'] = (-s/4, -h/4)
        self.anchors['NNE'] = (3*w/4+s/2, h/2)
        self.anchors['SSE'] = (3*w/4-s/2, -h/2)
        self.anchors['NNW'] = (w/4+s/2, h/2)
        self.anchors['SSW'] = (w/4-s/2, -h/2)


class Ellipse(Box):
    ''' Flowchart ellipse

        Args:
            w: Width of ellipse
            h: Height of ellipse

        Anchors:
            * 16 compass points (N, S, E, W, NE, NNE, etc.)
    '''
    def _set_segments(self):
        w, h = self.params['w'], self.params['h']
        # There's no ellipse Segment type, so draw one with a path Segment
        t = linspace(0, math.pi*2, num=50)
        y = [(h/2) * math.sin(t0) for t0 in t]
        x = [(w/2) * math.cos(t0) + w/2 for t0 in t]
        x[-1] = x[0]
        y[-1] = y[0]  # Ensure the path is actually closed
        self.segments.append(Segment(list(zip(x, y))))

    def _set_anchors(self):
        super()._set_anchors()
        w, h = self.params['w'], self.params['h']
        sinpi4 = math.sin(math.pi/4)
        cospi4 = math.cos(math.pi/4)
        sinpi8 = math.sin(math.pi/8)
        cospi8 = math.cos(math.pi/8)
        self.anchors['SE'] = (w/2+w/2*cospi4, -h/2*sinpi4)
        self.anchors['SW'] = (w/2-w/2*cospi4, -h/2*sinpi4)
        self.anchors['NW'] = (w/2-w/2*cospi4, h/2*sinpi4)
        self.anchors['NE'] = (w/2+w/2*cospi4, h/2*sinpi4)
        self.anchors['ENE'] = (w/2+w/2*cospi8, h/2*sinpi8)
        self.anchors['WNW'] = (w/2-w/2*cospi8, h/2*sinpi8)
        self.anchors['ESE'] = (w/2+w/2*cospi8, -h/2*sinpi8)
        self.anchors['WSW'] = (w/2-w/2*cospi8, -h/2*sinpi8)
        self.anchors['NNE'] = (w/2+w/2*sinpi8, h/2*cospi8)
        self.anchors['NNW'] = (w/2-w/2*sinpi8, h/2*cospi8)
        self.anchors['SSE'] = (w/2+w/2*sinpi8, -h/2*cospi8)
        self.anchors['SSW'] = (w/2-w/2*sinpi8, -h/2*cospi8)


class Decision(Box):
    ''' Flowchart decision (diamond)

        Args:
            w: Width of box
            h: Height of box
            N: text for North decision branch
            S: text for South decision branch
            E: text for East decision branch
            W: text for West decision branch
            font: Font family/name
            fontsize: Point size of label font

        Anchors:
            * 16 compass points (N, S, E, W, NE, NNE, etc.)
    '''
    def __init__(self, N: Optional[str] = None, E: Optional[str] = None, S: Optional[str] = None, W: Optional[str] = None,
                 font: Optional[str] = None, fontsize: float = 14,
                 **kwargs):
        super().__init__(**kwargs)
        self._size = 4, 2.25
        self._font = font
        self._fontsize = fontsize
        self._N = N
        self._E = E
        self._W = W
        self._S = S

    def _set_size(self) -> tuple[float, float]:
        minw, minh = self.params['minsize']
        pad = self.params['pad']
        mainlabels = [label for label in self._userlabels if label.loc is None]
        if mainlabels:
            minw, minh = labelsize(mainlabels[0], pad)
            tantheta = math.tan(math.radians(25)) # = h/2 / extraw
            fullw = minw + (minh/2/tantheta)
            fullh = minh + (tantheta*minw)
            minw = max(fullw, self.params['minsize'][0])
            minh = max(fullh, self.params['minsize'][1])

        self._userparams.setdefault('w', minw)
        self._userparams.setdefault('h', minh)
        return self._userparams['w'], self._userparams['h']
        
    def _set_segments(self):
        w, h = self.params['w'], self.params['h']
        font = self._font
        fontsize = self._fontsize
        self.segments.append(SegmentPoly([(0, 0), (w/2, h/2), (w, 0), (w/2, -h/2)]))

        lblofst = .13
        N, E, S, W = self._N, self._E, self._S, self._W
        if N:
            self.segments.append(SegmentText(
                (w/2+lblofst, h/2+lblofst), N, align=('left', 'bottom'),
                font=font, fontsize=fontsize))
        if S:
            self.segments.append(SegmentText(
                (w/2+lblofst, -h/2-lblofst), S, align=('left', 'top'),
                font=font, fontsize=fontsize))
        if E:
            self.segments.append(SegmentText(
                (w+lblofst, lblofst), E, align=('left', 'bottom'),
                font=font, fontsize=fontsize))
        if W:
            self.segments.append(SegmentText(
                (-lblofst, lblofst), W, align=('right', 'bottom'),
                font=font, fontsize=fontsize))

    def _set_anchors(self):
        w, h = self.params['w'], self.params['h']
        super()._set_anchors()
        self.anchors['NE'] = (3*w/4, h/4)
        self.anchors['NW'] = (w/4, h/4)
        self.anchors['SE'] = (3*w/4, -h/4)
        self.anchors['SW'] = (w/4, -h/4)
        self.anchors['NNE'] = (5*w/8, 3*h/8)
        self.anchors['NNW'] = (3*w/8, 3*h/8)
        self.anchors['SSE'] = (5*w/8, -3*h/8)
        self.anchors['SSW'] = (3*w/8, -3*h/8)
        self.anchors['ENE'] = (7*w/8, h/8)
        self.anchors['ESE'] = (7*w/8, -h/8)
        self.anchors['WNW'] = (w/8, h/8)
        self.anchors['WSW'] = (w/8, -h/8)


class Connect(Box):
    ''' Flowchart connector/circle

        Args:
            r: Radius of circle

        Anchors:
            * 16 compass points (N, S, E, W, NE, NNE, etc.)
    '''
    _element_defaults = {
        'h': 1.5,
        'w': 1.5,
        'r': 0.75,
        'minsize': (1.5, 1.5)
    }
    def _set_size(self) -> tuple[float, float]:
        pad = self.params['pad']
        mainlabels = [label for label in self._userlabels if label.loc is None]        
        if mainlabels:
            w, h = labelsize(mainlabels[0], pad)
            w = max(self.params['minsize'][0], w)
            h = max(self.params['minsize'][1], h)
        else:
            w, h = self.params['minsize']

        if 'r' in self._userparams:
            w = h = self._userparams.get('r')*2
        self._userparams.setdefault('w', w)
        self._userparams.setdefault('h', h)
        return self._userparams['w'], self._userparams['h']

    def _set_segments(self):
        r2 = self.params['w']
        r = r2/2
        self.segments = []
        self.segments.append(SegmentCircle((r, 0), r))

    def _set_anchors(self):
        super()._set_anchors()
        r2 = self.params['w']
        r = r2/2
        rsqrt2 = r * math.sqrt(2) / 2
        self.anchors['SE'] = (r+rsqrt2, -rsqrt2)
        self.anchors['SW'] = (r-rsqrt2, -rsqrt2)
        self.anchors['NE'] = (r+rsqrt2, rsqrt2)
        self.anchors['NW'] = (r-rsqrt2, rsqrt2)
        r225 = r * math.cos(math.radians(22.5))
        r675 = r * math.cos(math.radians(67.5))
        self.anchors['NNE'] = (r+r675, r225)
        self.anchors['ENE'] = (r+r225, r675)
        self.anchors['ESE'] = (r+r225, -r675)
        self.anchors['SSE'] = (r+r675, -r225)
        self.anchors['NNW'] = (r-r675, r225)
        self.anchors['WNW'] = (r-r225, r675)
        self.anchors['WSW'] = (r-r225, -r675)
        self.anchors['SSW'] = (r-r675, -r225)


class StateEnd(Connect):
    ''' End/Accept State (double circle)

        Args:
            r: radius
            dr: distance between circles

        Anchors:
            * 16 compass points (N, S, E, W, NE, NNE, etc.)
    '''
    def _set_size(self) -> tuple[float, float]:
        w, h = super()._set_size()
        dr = self._userparams.get('dr', .15)
        self.elmparams['w'] = w+dr*2
        self.elmparams['h'] = h+dr*2
        return self.params['w'], self.params['h']
        
    def _set_segments(self):
        super()._set_segments()
        r2 = self.params['w']
        r = r2/2
        dr = self._userparams.get('dr', .15)
        self.segments.append(SegmentCircle((r, 0), r-dr))


Process = Box
RoundProcess = RoundBox
Start = Terminal
Circle = Connect
State = Connect



from .pictorial import (
    INCH,
    MILLIMETER,
    PINSPACING,
    ElementPictorial,
    Resistor,
    LED,
    LEDBlue,
    LEDGreen,
    LEDOrange,
    LEDYellow,
    LEDWhite,
    CapacitorElectrolytic,
    CapacitorMylar,
    CapacitorCeramic,
    Diode,
    TO92,
    DIP,
    Breadboard)
from .fritz import FritzingPart, fritz_parts



''' Drawing Stack

    This module tracks the active drawing and element for use in context
    managers, enabling of adding elements to a drawing without explicitly
    calling `d.add` or `d +=`. The stack is a dictionary (relying on being
    ordered) of drawing: element pairs. The last item in the dictionary
    is the last drawing or container to be opened by `with` block, and
    the element is the last element instantiated.
    
    Typical order of operations:

    1) A drawing is added to the stack when the `with` block is opened
    2) When an element is instantiated inside the block, it is set as
        the value of the drawing's dictionary item.
    3) When a different element is instantiated, the element in the
        stack will then be added to its drawing if it hasn't been added already.
        This allows the chained methods (such as `.up`, `.down`) to affect
        the element before it is placed.
    4) When the `with` block exits, the last element to be instantiated
        is added to the drawing, if it hasn't been already. The drawing
        is then popped from the stack.

    A `pause` attribute may be set True to prevent any stack operations.
    This may be used, for example, when adding elements to an ElementCompound.
'''
from __future__ import annotations
from typing import Union, Optional, TYPE_CHECKING

if TYPE_CHECKING:
    from .schemdraw import Drawing
    from .elements import Element, Container

DrawingType = Union['Drawing', 'Container']
    
drawing_stack: dict[DrawingType, Optional['Element']] = {}  #{drawing: element}
pause: bool = False


def push_drawing(drawing: DrawingType) -> None:
    ''' Add a drawing to the stack '''
    drawing_stack[drawing] = None

def pop_drawing(drawing: DrawingType) -> None:
    ''' Remove the drawing from the stack '''
    drawing_stack.pop(drawing)
    
def push_element(element: Optional['Element']) -> None:
    ''' Add a new element to the stack, placing the existing
        one if not already placed by the user
    '''
    if not pause and len(drawing_stack) > 0:
        drawing, prev_elm = list(drawing_stack.items())[-1]
        if prev_elm is not None and prev_elm not in drawing:
            drawing.add(prev_elm)
        drawing_stack[drawing] = element




''' CSS/SVG Unit Conversions per
    https://www.w3.org/TR/css-values/#absolute-lengths
'''
PT_PER_IN = 72.
PX_PER_IN = 96.
CM_PER_IN = 2.54
MM_PER_IN = CM_PER_IN * 10
PC_PER_IN = 6.

PX_PER_PT = PX_PER_IN / PT_PER_IN

TO_PX = {
    'in': PX_PER_IN,
    'cm': PX_PER_IN / CM_PER_IN,
    'mm': PX_PER_IN / MM_PER_IN,
    'pt': PX_PER_IN / PT_PER_IN,
    'pc': PX_PER_IN / PC_PER_IN,
    'px': 1,
}


def parse_size_to_px(value: str) -> float:
    ''' Convert SVG size string (such as `2in`) to pixels '''
    try:
        value_px = float(value)
    except ValueError:
        unit = value[-2:]
        value_px = float(value[:-2]) * TO_PX.get(unit, 1)
    return value_px



''' Timing Diagrams, based on WaveJSON format '''

from __future__ import annotations
import re
import io
import ast
import tokenize
import math
from collections import namedtuple, ChainMap

from ..elements import Element
from ..segments import Segment, SegmentText, SegmentPoly, SegmentBezier
from ..backends.svg import text_size
from ..types import BBox
from ..util import Point
from .timingwaves import Wave0, Wave1, WaveL, WaveH, Wavez, WaveV, WaveU, WaveD, WaveClk, getsplit

LabelInfo = namedtuple('LabelInfo', ['name', 'row', 'height', 'level'])

PTS_TO_UNITS = 2/72


def state_level(state: str, ud: bool = False, np: bool = False) -> str:
    ''' Get level of wave state (0, 1, V, z, -) '''
    state = re.sub(r'[2-9]|\=|x', 'V', state)
    state = re.sub('u', '1', state)
    state = re.sub('d', '0', state)
    state = re.sub('n|N', '1', state)
    state = re.sub('p|P', '0', state)
    state = re.sub('l|L', '0', state)
    state = re.sub('h|H', '1', state)
    return state


def flatten(S):
    ''' Flatten the signals list so only individual rows (dicts) are included '''
    if S == []:
        return S
    if isinstance(S[0], list):
        return flatten(S[0]) + flatten(S[1:])
    if isinstance(S[0], str):
        return flatten(S[1:])
    return S[:1] + flatten(S[1:])


def max_depth(sig) -> int:
    ''' Get maximum group depth of the signal list '''
    return isinstance(sig, list) and max(map(max_depth, sig))+1


def get_nrows(sig) -> int:
    ''' Get the number of signal rows in the signal list '''
    if isinstance(sig, dict):
        return 1
    elif isinstance(sig[0], list):
        return sum(map(get_nrows, sig[0]))
    elif isinstance(sig[0], str):
        return sum(map(get_nrows, sig[1:]))
    return 0


def getlabels(sig, row: int = 0, level: int = 0) -> list[LabelInfo]:
    ''' Get a list of group label info '''
    if isinstance(sig, dict) or sig == []:
        return []
    elif all(isinstance(s, dict) for s in sig):
        return []
    elif isinstance(sig[0], str):
        lbl = [LabelInfo(sig[0], row, get_nrows(sig), level)]
        n = 0
        for s in sig[1:]:
            lbl.extend(getlabels(s, row=row+n, level=level+1))
            n += get_nrows(s)
        return lbl
    else:
        lbl = []
        n = 0
        for s in sig:
            lbl.extend(getlabels(s, row=row+n, level=level+1))
            n += get_nrows(s)
        return lbl


class TimingDiagram(Element):
    ''' Logic Timing Diagram

        Draw timing diagrams compatible with WaveJSON format
        See https://wavedrom.com/ for details. Use `from_json`
        to use WaveJSON strings copied from the site (since they
        can't be copied as proper Python dicts due to lack of
        quoting).

        Schemdraw provides a few additional extensions to the
        WaveJSON dictionary, including asynchronous waveforms
        and configuration options (color, lw) on each wave.
        See documentation for full specification.

        Args:
            wave: WaveJSON as a Python dict

        Keyword Args:
            yheight: Height of one waveform
            ygap: Separation between two waveforms
            risetime: Rise/fall time for wave transitions
            fontsize: Size of label fonts
            nodesize: Size of node labels
            namecolor: Color for wave names
            datacolor: Color for wave data text
            nodecolor: Color for node text
            gridcolor: Color of background grid
    '''
    _wavelookup = {'0': Wave0,
                   '1': Wave1,
                   'H': WaveH,
                   'h': WaveH,
                   'L': WaveL,
                   'l': WaveL,
                   'z': Wavez,
                   'u': WaveU,
                   'd': WaveD,
                   'n': WaveClk,
                   'p': WaveClk,
                   'N': WaveClk,
                   'P': WaveClk,
                   }
    _element_defaults = {
        'yheight': 0.5,
        'ygap': 0.3,
        'risetime': 0.15,
        'fontsize': 12,
        'nodesize': 8,
        'namecolor': 'blue',
        'datacolor': None,  # Inherit
        'nodecolor': None,  # Inherit
        'gridcolor': '#DDDDDD',
        'edgecolor': 'blue',
        'grid': True,
    }
    def __init__(self, waved: dict[str, str], **kwargs):
        super().__init__(**kwargs)
        self.wave = waved
        self.yheight = self.params['yheight']
        self.ygap = self.params['ygap']
        self.risetime = self.params['risetime']
        self.fontsize = self.params['fontsize']
        self.nodesize = self.params['nodesize']
        self.namecolor = self.params['namecolor']
        self.datacolor = self.params['datacolor']  # default: get color from theme
        self.nodecolor = self.params['nodecolor']
        self.gridcolor = self.params['gridcolor']
        self.edgecolor = self.params['edgecolor']
        self.grid = self.params['grid']
        self.kwargs = kwargs

        signals = self.wave.get('signal', [])  # type: ignore
        signals_flat = flatten(signals)
        config = self.wave.get('config', {})  # type: ignore
        self.hscale = config.get('hscale', 1)  # type: ignore

        height = (self.yheight+self.ygap)*len(signals_flat)
        periods = max(len(w.get('wave', [])) for w in signals_flat)
        periods = max(periods, (max(w.get('async', [0])[-1] for w in signals_flat)))
        if self.grid:
            self._drawgrid(periods, height)

        # phase shifts that go off screen will be clipped by this rect
        # +.05 so the top pixel row doesn't get clipped
        clipbox = BBox(0, self.yheight+.05, periods*self.yheight*2*self.hscale, -height)
        self.kwargs['clip'] = clipbox

        labelwidth = 0.
        y0 = 0.
        for signal in signals_flat:
            name = signal.get('name', '')
            _width = self._drawname(name, y0)
            labelwidth = max(labelwidth, _width)

            if 'async' in signal:
                self._drawasync(signal, y0)
            else:
                self._drawwave(signal, y0=y0)
            self._drawnodes(signal, y0=y0)
            y0 -= (self.yheight+self.ygap)

        self._drawedges()
        self._drawgroups(signals, labelwidth)

    def _drawgrid(self, periods, height):
        ''' Draw grid (vertical dotted lines) '''
        for p in range(periods+1):
            self.segments.append(
                Segment([(p*2*self.yheight*self.hscale, self.yheight+self.ygap/2),
                         (p*2*self.yheight*self.hscale, self.yheight-height)],
                        ls=':', lw=1, color=self.gridcolor, zorder=0))

    def _drawname(self, name, y0):
        ''' Draw name of one wave. Returns calculated unit width of the string. '''
        textpad = .2
        self.segments.append(
            SegmentText((-textpad, y0), name, align=('right', 'bottom'),
                        fontsize=self.fontsize, color=self.namecolor))
        return text_size(name, size=self.fontsize)[0] * PTS_TO_UNITS

    def _drawwave(self, signal, y0=0):
        ''' Draw one wave.

            Args:
                signal: Dictionary defining the signal to draw
                y0: Vertical position
        '''
        wave = signal.get('wave', '')
        phase = signal.get('phase', 0)
        waverise = signal.get('risetime', None)
        wavekwargs = ChainMap({'color': signal.get('color', None),
                               'lw': signal.get('lw', 1),
                               'clip': self.kwargs.get('clip')})
        data = signal.get('data', [])
        if not isinstance(data, list):
            data = data.split()  # Sometimes it's a space-separated string...

        period = 2*self.yheight*signal.get('period', 1) * self.hscale
        y1 = y0 + self.yheight
        i = 0
        pstate = '-'

        x = -period*phase
        while i < len(wave):
            state = wave[i]
            splits = []
            periods = 1
            k = i+1
            while k < len(wave) and wave[k] in '|.':
                if wave[k] == '|':
                    splits.append(periods)
                periods += 1
                k += 1
            nstate = wave[k] if k < len(wave) else '-'

            xend = x+periods*period
            params = {'state': state,
                      'pstate': pstate,
                      'nstate': nstate,
                      'plevel': state_level(pstate),
                      'nlevel': state_level(nstate),
                      'periods': periods,
                      'period': period,
                      'x0': x,
                      'xend': xend,
                      'y0': y0,
                      'y1': y1,
                      'rise': waverise if waverise is not None else self.risetime,
                      'data': data,
                      'datacolor': self.datacolor,
                      'kwargs': wavekwargs}

            wavecls = self._wavelookup.get(state, WaveV)
            self.segments.extend(wavecls(params).segments())

            for split in splits:
                self.segments.extend(
                    getsplit(x + (split+1)*period-period/2, y0, y1))

            pstate = state
            x += periods*period
            i = k

    def _drawasync(self, signal, y0):
        ''' Draw asynchronous wave

            Args:
                signal: Dictionary defining the signal to draw
                y0: Vertical position
        '''
        times = signal.get('async', '')
        wave = signal.get('wave', '')
        waverise = signal.get('risetime', None)
        wavekwargs = ChainMap({'color': signal.get('color', None),
                               'lw': signal.get('lw', 1), 
                               'clip': self.kwargs.get('clip')})
        rise = waverise if waverise is not None else self.risetime

        data = signal.get('data', [])
        if not isinstance(data, list):
            data = data.split()  # Sometimes it's a space-separated string...
        if not isinstance(times, list):
            times = [float(f) for f in times.split()]

        if len(wave) + 1 != len(times):
            raise ValueError('len(times) must be one more than len(wave).')

        period = 2*self.yheight*signal.get('period', 1) * self.hscale
        y1 = y0 + self.yheight
        pstate = '-'
        for i in range(len(wave)):
            state = wave[i]
            t0 = times[i]
            t1 = times[i+1]
            nstate = wave[i+1] if i<len(wave)-1 else '-'
            x = t0*period
            xend = t1*period
            params = {'state': state,
                      'pstate': pstate,
                      'nstate': nstate,
                      'plevel': state_level(pstate),
                      'nlevel': state_level(nstate),
                      'periods': 0,
                      'period': period,
                      'x0': x,
                      'xend': xend,
                      'y0': y0,
                      'y1': y1,
                      'rise': rise,
                      'data': data,
                      'datacolor': self.datacolor,
                      'kwargs': wavekwargs}

            wavecls = self._wavelookup.get(state, WaveV)
            self.segments.extend(wavecls(params).segments())
            x = xend
            pstate = state

    def _drawnodes(self, signal, y0):
        ''' Draw nodes (labels along the wave) and define anchors for each

            Args:
                signal: Dictionary defining the signal to draw
                y0: Vertical position
        '''
        nodes = signal.get('node', '')
        phase = signal.get('phase', 0)
        period = 2*self.yheight*signal.get('period', 1) * self.hscale

        y1 = y0 + self.yheight
        for j, node in enumerate(nodes):
            if node == '.':
                continue
            w, h, _ = text_size(node, size=self.nodesize)
            w, h = w*PTS_TO_UNITS*2.5, h*PTS_TO_UNITS*2.5
            ycenter = (y0+y1)/2
            xnode = j*period + self.risetime/2 - period*phase
            if not node.isupper():  # Only uppercase nodes and symbols are drawn
                self.segments.append(SegmentPoly([(xnode-w/2, ycenter-h/2), (xnode-w/2, ycenter+h/2),
                                                  (xnode+w/2, ycenter+h/2), (xnode+w/2, ycenter-h/2)],
                                                 color='none', fill='bg', zorder=3))
                self.segments.append(
                    SegmentText((xnode, ycenter), node, align=('center', 'center'),
                                fontsize=self.nodesize, color=self.nodecolor, zorder=3))
            self.anchors[f'node_{node}'] = (xnode, ycenter)

    def _drawedges(self):
        edges = self.wave.get('edge', [])  # type: ignore
        chrrad = self.nodesize / 60
        caplen = .1
        period = 2*self.yheight * self.hscale

        for edge in edges:
            label = ''
            if ' ' in edge:
                edge, label = edge.split(maxsplit=1)
            style = re.findall(r'{(.*)}', edge)
            edge = re.sub(r'{(.*)}', '', edge)
            nodes = re.findall(r'\[(.+?)\]', edge)
            if len(nodes) == 0:
                # Single letter node names, WaveDrom style
                mode = edge[1:-1]
                p0 = Point(self.anchors[f'node_{edge[0]}'])
                pn = Point(self.anchors[f'node_{edge[-1]}'])

            else:
                # Extended node naming - [WaveNumber:Xposition]
                assert len(nodes) == 2
                mode = re.subn(r'\[(.+?)\]', '', edge)[0]
                endpoints = []
                for node in nodes:
                    nodewave, nodet = node.split(':')
                    nodet = float(nodet)  # number periods into wave
                    ofst = self.yheight/2
                    if '^' in nodewave:
                        ofst = self.yheight + caplen*2
                    elif 'v' in nodewave:
                        ofst = -caplen*2
                    nodewave = nodewave.replace('^', '').replace('v', '')
                    wavenum = int(nodewave)
                    endpoints.append(
                        Point((nodet*period+self.risetime/2,
                               -wavenum*(self.yheight + self.ygap) + ofst)))
                p0, pn = endpoints

            color = self.edgecolor
            ls = '-'
            if style:
                style = style[0].split(',')
                color = style[0]
                ls = '-' if len(style) == 1 else style[1]

            if '<' in edge and '>' in edge:
                arrow = '<>'
            elif '<' in edge:
                arrow = '>'
            elif '>' in edge:
                arrow = '<'
            else:
                arrow = None

            mode = mode.strip('<>')
            center = Point(((p0.x+pn.x)/2, (p0.y+pn.y)/2))
            if mode == '-':  # Straight line
                center = Point(((p0.x+pn.x)/2, (p0.y+pn.y)/2))
                th0 = math.atan2((pn.y-p0.y), (pn.x-p0.x))
                p0 = Point((p0.x + chrrad * math.cos(th0), p0.y + chrrad * math.sin(th0)))
                pn = Point((pn.x - chrrad * math.cos(th0), pn.y - chrrad * math.sin(th0)))
                self.segments.append(Segment([pn, p0], lw=1, ls=ls, color=color,
                                             arrow=arrow, zorder=3))

            elif mode == '+':  # Straight line with endcaps, full length
                center = Point(((p0.x+pn.x)/2, (p0.y+pn.y)/2))
                th0 = math.atan2(-(pn.x-p0.x), (pn.y-p0.y))
                self.segments.append(Segment((p0, pn), lw=1, ls=ls, color=color))
                cap1 = [(p0.x + caplen*math.cos(th0), p0.y + caplen*math.sin(th0)),
                        (p0.x - caplen*math.cos(th0), p0.y - caplen*math.sin(th0))]
                cap2 = [(pn.x + caplen*math.cos(th0), pn.y + caplen*math.sin(th0)),
                        (pn.x - caplen*math.cos(th0), pn.y - caplen*math.sin(th0))]
                self.segments.append(Segment(cap1, lw=1, color=color))
                self.segments.append(Segment(cap2, lw=1, color=color))

            elif mode == '|-':
                center = Point((p0.x, (p0.y+pn.y)/2))
                dy = 1 if p0.y > pn.y else -1
                dx = 1 if p0.x < pn.x else -1
                p0 = p0 - Point((0, chrrad*dy))
                pn = pn - Point((chrrad*dx, 0))
                p1 = Point((p0.x, pn.y))
                self.segments.append(Segment((p0, p1, pn), lw=1, ls=ls, color=color, zorder=3, arrow=arrow))

            elif mode == '-|':
                center = Point((pn.x, (p0.y+pn.y)/2))
                dy = -1 if p0.y > pn.y else 1
                dx = -1 if p0.x < pn.x else 1
                p0 = p0 - Point((chrrad*dx, 0))
                pn = pn - Point((0, chrrad*dy))
                p1 = Point((pn.x, p0.y))
                self.segments.append(Segment((p0, p1, pn), lw=1, ls=ls, color=color, arrow=arrow, zorder=3))

            elif mode == '-|-':
                center = (p0+pn)/2
                dx = -1 if pn.x < p0.x else 1
                p0 = p0 + Point((chrrad*dx, 0))
                p1 = Point((center.x, p0.y))
                p2 = Point((center.x, pn.y))
                pn = pn - Point((chrrad*dx, 0))
                self.segments.append(Segment((p0, p1, p2, pn), lw=1, ls=ls, color=color, arrow=arrow, zorder=3))

            elif mode == '~':  # S-curve, start and end horizontally
                center = Point(((p0.x+pn.x)/2, (p0.y+pn.y)/2))
                p0 = p0 + Point((chrrad, 0))
                p3 = pn - Point((chrrad, 0))
                dx = p3.x - p0.x
                p1 = p0 + Point((dx*.6, 0))
                p2 = p3 - Point((dx*.6, 0))
                self.segments.append(SegmentBezier([p0, p1, p2, p3], lw=1, ls=ls,
                                                   color=color, arrow=arrow, zorder=3))

            elif mode == '-~':  # C curve, horizontal at start
                center = Point((p0.x+0.8*(pn.x-p0.x), p0.y + (pn.y-p0.y)/2))
                dx = pn.x - p0.x
                p1 = p0 + Point((dx*.7, 0))
                th0 = math.atan2((p1.y-p0.y), (p1.x-p0.x))
                th2 = math.atan2((p1.y-pn.y), (p1.x-pn.x))
                p0 = Point((p0.x + chrrad * math.cos(th0), p0.y + chrrad * math.sin(th0)))
                pn = Point((pn.x - chrrad * math.cos(th0), pn.y + chrrad * math.sin(th2)))
                self.segments.append(SegmentBezier(
                    [p0, p1, pn], lw=1, ls=ls, color=color, arrow=arrow, zorder=3))

            elif mode == '~-':
                center = Point((p0.x+0.25*(pn.x-p0.x), (p0.y+pn.y)/2))
                dx = pn.x - p0.x
                p1 = pn - Point((dx*.6, 0))
                th0 = math.atan2((p1.y-p0.y), (p1.x-p0.x))
                th2 = math.atan2((p1.y-pn.y), (p1.x-pn.x))
                p0 = Point((p0.x + chrrad * math.cos(th0), p0.y + chrrad * math.sin(th0)))
                pn = Point((pn.x + chrrad * math.cos(th2), pn.y + chrrad * math.sin(th2)))
                self.segments.append(SegmentBezier(
                    [p0, p1, pn], lw=1, ls=ls, color=color, arrow=arrow, zorder=3))

            if label:
                w, h, _ = text_size(label, size=self.nodesize)
                w, h = w*PTS_TO_UNITS*1.5, h*PTS_TO_UNITS*1.5
                self.segments.append(SegmentPoly([(center.x-w/2, center.y-h/2),
                                                  (center.x-w/2, center.y+h/2),
                                                  (center.x+w/2, center.y+h/2),
                                                  (center.x+w/2, center.y-h/2)],
                                     color='none', fill='bg', zorder=4))
                self.segments.append(SegmentText(center, label, fontsize=self.nodesize,
                                                 color=self.nodecolor, align=('center', 'center'),
                                                 zorder=4))

    def _drawgroups(self, signals, labelwidth):
        ''' Draw group labels

            Args:
                signals: List of dictionaries defining the signals
                labelwidth: Max text width (drawing units) of signal names
        '''
        nlevels = max_depth(signals)
        edgelen = .05
        levelwidth = 0.5
        for label in getlabels(signals):
            xval = -(labelwidth) - (nlevels-label.level)*levelwidth
            ytop = -label.row*(self.yheight+self.ygap) + self.yheight - edgelen
            ybot = ytop - label.height * (self.yheight+self.ygap) + self.ygap/2 + edgelen
            ycenter = (ytop+ybot)/2
            xtext = xval - .1
            self.segments.append(
                Segment([(xval+edgelen, ybot-edgelen), (xval, ybot),
                         (xval, ytop), (xval+edgelen, ytop+edgelen)],
                        color=self.namecolor, lw=1))
            self.segments.append(
                SegmentText((xtext, ycenter), label.name, rotation=90,
                            align=('center', 'bottom'), color=self.namecolor,
                            fontsize=self.fontsize))

    @classmethod
    def from_json(cls, wave: str, **kwargs):
        ''' Create timing diagram from string WaveJSON. '''
        # Source for cleaning up JSON: https://stackoverflow.com/a/61783377/13826284
        tokens = tokenize.generate_tokens(io.StringIO(wave).readline)
        modified_tokens = (
            (tokenize.STRING, repr(token.string)) if token.type == tokenize.NAME else token[:2]
            for token in tokens)

        fixed_input = tokenize.untokenize(modified_tokens)
        return cls(ast.literal_eval(fixed_input), **kwargs)



''' Logic gate definitions '''
from __future__ import annotations
from typing import Optional, Sequence
from functools import partial
import math

from ..segments import Segment, SegmentCircle
from ..elements import Element, Element2Term
from ..util import linspace


gap = (math.nan, math.nan)
gateh = 1.
gatel = .65
notbubble = .12


class And(Element):
    ''' AND gate

        Args:
            inputs: Number of inputs to gate.
            nand: Draw invert bubble on output
            inputnots: Input numbers (starting at 1) of inputs that have invert bubble
            leadin: Length of input leads [default: 0.35]
            leadout: Length of output lead [default: 0.35]

        Anchors:
            out
            in[X] - for each input
    '''
    _element_defaults = {
        'leadin': 0.35,
        'leadout': 0.35
    }
    def __init__(self, inputs: int = 2, nand: bool = False, inputnots: Optional[Sequence[int]] = None,
                 *,
                 leadin: Optional[float] = None, leadout: Optional[float] = None, **kwargs):
        super().__init__(**kwargs)
        _leadin = self.params['leadin']
        _leadout = self.params['leadout']
        rad = gateh/2
        theta = linspace(math.radians(-90), math.radians(90), num=50)
        xs = [rad*math.cos(t) + gatel+_leadin for t in theta]
        ys = [rad*math.sin(t) for t in theta]

        path = list(zip(xs, ys))
        path += [(gatel+_leadin, rad), (_leadin, rad),
                 (_leadin, 0), (_leadin, -rad),
                 (gatel+_leadin, -rad)]
        self.segments.append(Segment(path))
        self.anchors['out'] = (gatel+gateh/2+_leadin+_leadout, 0)
        self.anchors['end'] = self.anchors['out']

        if nand:
            self.segments.append(SegmentCircle(
                (_leadin+gatel+gateh/2+notbubble, 0), notbubble))

        # Set distance between inputs. A little larger for 2 input gates.
        if inputs == 2:
            dy = gateh*.5
        elif inputs == 3:
            dy = gateh*.33
        else:  # inputs > 3:
            dy = gateh*.4
            backlen = dy * (inputs-1)

        # Add the inputs and define anchor names
        for i in range(inputs):
            y = (i+1 - (inputs/2+.5)) * dy
            self.anchors['in%d' % (inputs-i)] = (0, y)

            if inputnots and (inputs-i) in inputnots:
                self.segments.append(SegmentCircle((_leadin-notbubble, y), notbubble))
                self.segments.append(Segment([(0, y), (_leadin-notbubble*2, y)]))
            else:
                self.segments.append(Segment([(0, y), (_leadin, y)]))

        # Extended back for large number of inputs
        if inputs > 3:
            self.segments.append(Segment([(_leadin, backlen/2+dy/2),
                                          (_leadin, -backlen/2-dy/2)]))

        # Output lead
        if nand:
            self.segments.append(Segment(
                [(gatel+gateh/2+_leadin+notbubble*2, 0), (gatel+gateh/2+_leadin+_leadout, 0)]))
        else:
            self.segments.append(Segment(
                [(gatel+gateh/2+_leadin, 0), (gatel+gateh/2+_leadin+_leadout, 0)]))
        self.params['drop'] = self.segments[-1].path[-1]  # type: ignore


Nand = partial(And, nand=True)


class Or(Element):
    ''' OR or XOR gate element.

        Args:
            inputs: Number of inputs to gate.
            nor: Draw invert bubble on output
            xor: Draw as exclusive-or gate
            inputnots: Input numbers (starting at 1) of inputs that have invert bubble
            leadin: Length of input leads [default: 0.35]
            leadout: Length of output lead [default: 0.35]

        Anchors:
            out
            in[X] - for each input
    '''
    _element_defaults = {
        'leadin': 0.35,
        'leadout': 0.35
    }
    def __init__(self, inputs: int = 2, nor: bool = False,
                 xor: bool = False, inputnots: Optional[Sequence[int]] = None,
                 *,
                 leadin: float = 0.35, leadout: float = 0.35, **kwargs):
        super().__init__(**kwargs)
        _leadin = self.params['leadin']
        _leadout = self.params['leadout']
        # Define OR path
        orflat = .5
        xorgap = .15
        xs = linspace(0, gatel+.05)
        ys = [x0**2 for x0 in xs]
        ys = [y0 - max(ys) for y0 in ys]
        ys = [min(ys)] + ys   # Combine the flat + parabolic parts
        xs = [0.] + [x0+orflat for x0 in xs]

        # Back/input side
        y2 = linspace(min(ys), -min(ys))
        x2 = [-y0**2 for y0 in y2]
        back = min(x2)
        x2 = [x0 - back for x0 in x2]

        # Offset for inputs
        xs = [x0 + _leadin for x0 in xs]
        x2 = [x0 + _leadin for x0 in x2]

        tip = max(xs)
        orheight = abs(min(ys))

        negy = [-y0 for y0 in ys]
        path = list(zip(xs, ys)) + list(zip(xs[::-1], negy[::-1]))
        path += list(zip(x2[::-1], y2[::-1]))

        self.segments.append(Segment(path))
        self.anchors['out'] = (tip+_leadout, 0)
        self.anchors['end'] = self.anchors['out']

        if xor:
            xxor = [x0-xorgap for x0 in x2]
            self.segments.append(Segment(list(zip(xxor, y2))))
            _leadin -= xorgap

        if nor:
            self.segments.append(SegmentCircle((tip+notbubble, 0), notbubble))

        # Set distance between inputs. A little larger for 2 input gates.
        if inputs == 2:
            dy = gateh*.5
        elif inputs == 3:
            dy = gateh*.33
        else:  # inputs > 3:
            dy = gateh*.4
            backlen = dy * (inputs-1)

        # Add the inputs and define anchor names
        for i in range(inputs):
            y = (i+1 - (inputs/2+.5)) * dy

            xback = _leadin - y**2 - back
            if inputs > 3 and ((y > orheight) or (y < -orheight)):
                xback = _leadin

            self.anchors['in%d' % (inputs-i)] = (0, y)

            if inputnots and (inputs-i) in inputnots:
                self.segments.append(SegmentCircle(
                    (xback-notbubble, y), notbubble))
                self.segments.append(Segment([(0, y), (xback-notbubble*2, y)]))
            else:
                self.segments.append(Segment([(0, y), (xback, y)]))

        # Extended back for large number of inputs
        if inputs > 3:
            self.segments.append(Segment(
                [(_leadin, backlen/2+dy/2), (_leadin, orheight)]))
            self.segments.append(Segment(
                [(_leadin, -backlen/2-dy/2), (_leadin, -orheight)]))

        # Output lead
        if nor:
            self.segments.append(Segment([(tip+notbubble*2, 0), (tip+_leadout, 0)]))
        else:
            self.segments.append(Segment([(tip, 0), (tip+_leadout, 0)]))
        self.elmparams['drop'] = self.segments[-1].path[-1]  # type: ignore
        self.tip = tip


Nor = partial(Or, nor=True)
Xor = partial(Or, xor=True)
Xnor = partial(Or, nor=True, xor=True)


class Buf(Element2Term):
    ''' Buffer

        Anchors:
            in
            out
    '''
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.segments.append(Segment([(0, 0), gap, (gatel, 0)]))
        self.segments.append(Segment([(0, gateh/2), (gatel, 0), (0, -gateh/2), (0, gateh/2)]))
        self.anchors['out'] = (gatel, 0)
        self.anchors['in1'] = (0, 0)


class Not(Element2Term):
    ''' Not gate/inverter

        Anchors:
            in
            out
    '''
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.segments.append(Segment([(0, 0), gap, (gatel+notbubble*2, 0)]))
        self.segments.append(Segment([(0, gateh/2), (gatel, 0), (0, -gateh/2), (0, gateh/2)]))
        self.segments.append(SegmentCircle((gatel+notbubble, 0), notbubble))
        self.anchors['out'] = (gatel+notbubble*2, 0)
        self.anchors['in1'] = (0, 0)


class Tristate(Element2Term):
    ''' Tristate inverter
        Args:
            outputnot: Draw invert bubble on output
            controlnot: Active-low control

        Anchors:
            in
            out
            c
    '''
    def __init__(self, outputnot: bool = True, controlnot: bool = False, **kwargs):
        kwargs['outputnot'] = outputnot
        kwargs['control'] = controlnot
        super().__init__(**kwargs)
        if outputnot:
          self.segments.append(Segment([(0, 0), gap, (gatel+notbubble*2, 0)]))
          self.segments.append(SegmentCircle((gatel+notbubble, 0), notbubble))
        else:
          self.segments.append(Segment([(0, 0), gap, (gatel, 0)]))
        self.segments.append(Segment([(0, gateh/2), (gatel, 0), (0, -gateh/2), (0, gateh/2)]))
        if controlnot:
          self.segments.append(SegmentCircle((gatel/2, .42), notbubble))
          self.segments.append(Segment([(gatel/2, .42+notbubble), (gatel/2, .7)]))
        else:  
          self.segments.append(Segment([(gatel/2, .28), (gatel/2, .7)]))
        self.anchors['out'] = (gatel+notbubble*2, 0)
        self.anchors['in1'] = (0, 0)
        self.anchors['c'] = (gatel/2, .7)


class NotNot(Element2Term):
    ''' Double inverter

        Anchors:
            in
            out
    '''
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.segments.append(Segment([(-notbubble*2, 0), gap, (gatel+notbubble*2, 0)]))
        self.segments.append(Segment([(0, gateh/2), (gatel, 0), (0, -gateh/2), (0, gateh/2)]))
        self.segments.append(SegmentCircle((gatel+notbubble, 0), notbubble))
        self.segments.append(SegmentCircle((gatel+notbubble, 0), notbubble))
        self.segments.append(SegmentCircle((-notbubble, 0), notbubble))
        self.anchors['out'] = (gatel+notbubble*2, 0)
        self.anchors['in1'] = (0, 0)


class Tgate(Element2Term):
    ''' Transmission gate.

        Anchors:
            in
            out
            c
            cbar
    '''
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.segments.append(Segment(
            [(0, 0), gap, (gatel, 0)]))
        self.segments.append(Segment(
            [(0, 0), (0, -gateh/2), (gatel, 0),
             (0, gateh/2), (0, 0)]))
        self.segments.append(Segment(
            [(gatel, 0), (gatel, -gateh/2), (0, 0),
             (gatel, gateh/2), (gatel, 0)]))
        self.segments.append(SegmentCircle((gatel/2, .28+.08), .08))
        self.segments.append(Segment(
            [(gatel/2, .28+.16), (gatel/2, .7)]))
        self.segments.append(Segment(
            [(gatel/2, -.28), (gatel/2, -.7)]))
        self.anchors['out'] = (gatel, 0)
        self.anchors['in1'] = (0, 0)
        self.anchors['c'] = (gatel/2, -.7)
        self.anchors['cbar'] = (gatel/2, .7)


class Schmitt(Buf):
    ''' Schmitt Trigger

        Anchors:
            in
            out
    '''
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.segments.append(Segment([(.07, -.15), (.25, -.15), (.25, .15), gap,
                                      (.33, .15), (.15, .15), (.15, -.15)], lw=1))


class SchmittNot(Not):
    ''' Inverted Schmitt Trigger

        Anchors:
            in
            out
    '''
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.segments.append(Segment([(.07, -.15), (.25, -.15), (.25, .15), gap,
                                      (.33, .15), (.15, .15), (.15, -.15)], lw=1))


class SchmittAnd(And):
    ''' Schmitt Trigger AND

        Anchors:
            in1
            in2
            out
    '''
    def __init__(self, *, leadin: float = 0.35, leadout: float = 0.35, **kwargs):
        super().__init__(leadin=leadin, leadout=leadout, **kwargs)
        xofst = self.params['leadin']+gatel/2
        self.segments.append(Segment([(xofst+.07, -.15), (xofst+.25, -.15),
                                      (xofst+.25, .15), gap,
                                      (xofst+.33, .15), (xofst+.15, .15),
                                      (xofst+.15, -.15)], lw=1))


SchmittNand = partial(SchmittAnd, nand=True)
